import logging
import os

from dbnd._core.utils.basics.nothing import NOTHING, is_not_defined
from targets.file_target import FileTarget
from targets.fs import FileSystems
from targets.target_factory import target
from targets.utils.open_multiple import MultiTargetOpen


logger = logging.getLogger(__name__)

DEFAULT_FLAG_FILE_NAME = "_SUCCESS"


class DirTarget(FileTarget):
    def __init__(self, path, fs, config=None, io_pipe=None, source=None):
        """
         If `flag` Defines a target directory with a flag-file (defaults to `_SUCCESS`) used
    to signify job success.

        This checks for two things:

        * the path exists (just like the S3Target)
        * the _SUCCESS file exists within the directory.

        Because Hadoop outputs into a directory and not a single file,
        the path is assumed to be a directory.

        :param path:
        :param fs:
        :param config:
        :param io_pipe:
        :param flag:
        :param source:
        """

        super(DirTarget, self).__init__(
            path=path, fs=fs, config=config, io_pipe=io_pipe, source=source
        )
        if path[-1] != "/" and path[-1] != "\\":
            raise ValueError(
                "%s requires the path to be to a "
                "directory.  It must end with a slash ( / )." % self.__class__.__name__
            )
        flag = config.flag
        if flag is True:
            flag = DEFAULT_FLAG_FILE_NAME  # default value, otherwise override it

        self.flag_target = target(path, flag, fs=fs) if flag else None

        self._auto_partition_count = 0
        self._write_target = self.partition("part-0000")

        self.meta_files = config.meta_files

    def exists(self):
        path = self.path
        if self.flag_target:
            return self.flag_target.exists()

        return super(DirTarget, self).exists()

    def folder_exists(self):
        return self.fs.exists(self.path)

    def is_meta_file(self, path):
        base_name = os.path.basename(path)
        if base_name.startswith(".") or base_name.startswith("_"):
            return True
        if base_name in self.meta_files:
            return True
        return False

    def list_partitions(self):
        all_files = self.fs.listdir(self.path)
        all_files = sorted(all_files)
        all_files = filter(lambda x: not self.is_meta_file(x), all_files)
        if self.flag_target:
            all_files = filter(lambda x: x != str(self.flag_target), all_files)
        return [
            target(
                p,
                fs=self._fs,
                config=self.config.as_file(),
                properties=self.properties,
                source=self.source,
            )
            for p in all_files
        ]

    def open(self, mode="r"):
        if "r" in mode:
            return MultiTargetOpen(targets=self.list_partitions(), mode=mode)
        elif "w" in mode:
            return self._write_target.open(mode)
        else:
            raise ValueError("Unsupported open mode '{}'".format(mode))

    def partition(self, name=NOTHING, extension=NOTHING, config=NOTHING, **kwargs):
        """
        :param config:
        :param name: file name of the partition. if not provided - "part-%04d" % ID
        :param extension: extension. if not provided -> default extension will be used
        :return: FileTarget that represents the partition.
        """
        if is_not_defined(name):
            name = "part-%04d" % self._auto_partition_count
            self._auto_partition_count += 1
        if is_not_defined(config):
            # only if it's a call not from file,folder - we set it as file
            config = self.config.as_file()

        if is_not_defined(extension):
            extension = config.get_ext()
        if extension:
            name += extension
        return target(self.path, name, config=config, fs=self._fs, **kwargs)

    def mkdir(self):
        self.fs.mkdir(self.path)

    def mark_success(self):
        self.mkdir()
        if self.flag_target:
            self.flag_target.touch()

    def is_local(self):
        return self.fs_name == FileSystems.local

    def move_from(self, from_path, raise_if_exists=False):
        self.fs.move_dir(from_path, self.path, raise_if_exists)
        self.mark_success()

    def make_tmp(self):
        return self.fs.make_tmp_dir()



import logging
import time
import typing

from dbnd._core.constants import DbndTargetOperationType
from dbnd._core.plugin.dbnd_plugins import is_plugin_enabled
from dbnd._core.task_run.task_run_tracker import TaskRunTracker
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.utils import seven
from targets.value_meta import ValueMetaConf


if typing.TYPE_CHECKING:
    from datetime import datetime
    from typing import Optional, Union, List, Dict, Any
    import pandas as pd
    import pyspark.sql as spark

    from dbnd_postgres.postgres_values import PostgresTable
    from dbnd_snowflake.snowflake_values import SnowflakeTable

logger = logging.getLogger(__name__)


def _get_tracker():
    # type: ()-> Optional[TaskRunTracker]
    from dbnd._core.task_run.current_task_run import try_get_or_create_task_run

    task_run = try_get_or_create_task_run()
    if task_run:
        return task_run.tracker
    return None


def log_data(
    key,  # type: str
    value=None,  # type: Union[pd.DataFrame, spark.DataFrame, PostgresTable, SnowflakeTable]
    path=None,  # type: Optional[str]
    operation_type=DbndTargetOperationType.read,  # type: DbndTargetOperationType
    with_preview=None,  # type: Optional[bool]
    with_size=None,  # type: Optional[bool]
    with_schema=None,  # type: Optional[bool]
    with_stats=None,  # type: Optional[Union[bool, str, List[str], LogDataRequest]]
    with_histograms=None,  # type: Optional[Union[bool, str, List[str], LogDataRequest]]
    raise_on_error=False,  # type: bool
):  # type: (...) -> None
    tracker = _get_tracker()
    if not tracker:
        return

    meta_conf = ValueMetaConf(
        log_preview=with_preview,
        log_schema=with_schema,
        log_size=with_size,
        log_stats=with_stats,
        log_histograms=with_histograms,
    )

    tracker.log_data(
        key,
        value,
        meta_conf=meta_conf,
        path=path,
        operation_type=operation_type,
        raise_on_error=raise_on_error,
    )


log_dataframe = log_data


def log_pg_table(
    table_name,
    connection_string,
    with_preview=None,  # type: Optional[bool]
    with_schema=None,  # type: Optional[bool]
    with_histograms=None,  # type: Union[LogDataRequest, bool, str, List[str]]
):

    try:
        if not is_plugin_enabled("dbnd-postgres", module_import="dbnd_postgres"):
            return
        from dbnd_postgres import postgres_values

        pg_table = postgres_values.PostgresTable(table_name, connection_string)
        log_data(
            table_name,
            pg_table,
            with_preview=with_preview,
            with_schema=with_schema,
            with_histograms=with_histograms,
        )
    except Exception:
        logger.exception("Failed to log_pg_table")


def log_metric(key, value, source="user"):
    tracker = _get_tracker()
    if tracker:
        tracker.log_metric(key, value, source=source)
        return

    logger.info("Log {} Metric '{}'='{}'".format(source.capitalize(), key, value))


def log_metrics(metrics_dict, source="user", timestamp=None):
    # type: (Dict[str, Any], str, datetime) -> None
    tracker = _get_tracker()
    if tracker:
        from dbnd._core.tracking.schemas.metrics import Metric
        from dbnd._core.utils.timezone import utcnow

        metrics = [
            Metric(key=key, value=value, source=source, timestamp=timestamp or utcnow())
            for key, value in metrics_dict.items()
        ]
        tracker._log_metrics(metrics)


def log_artifact(key, artifact):
    tracker = _get_tracker()
    if tracker:
        tracker.log_artifact(key, artifact)
        return

    logger.info("Artifact %s=%s", key, artifact)


@seven.contextlib.contextmanager
def log_duration(metric_key, source="user"):
    """
    Measure time of function or code block, and log to Databand as a metric.
    Can be used as a decorator and in "with" statement as a context manager.

    Example 1:
        @log_duration("f_time_duration")
        def f():
            sleep(1)

    Example 2:
        with log_duration("my_code_duration"):
            sleep(1)
    """
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        log_metric(metric_key, end_time - start_time, source)



import datetime
import logging
import random

from dbnd._core.commands import log_metric
from dbnd._core.decorator.dbnd_decorator import task


logger = logging.getLogger(__name__)


@task
def dbnd_sanity_check(check_time=datetime.datetime.now()):
    # type: ( datetime.datetime)-> str
    logger.info("Running Sanity Check!")
    log_metric("Happiness Level", "High")
    logger.info("Your system is good to go! Enjoy Databand!")
    return "Databand checked at %s" % check_time


@task
def dbnd_random_check(check_time=datetime.datetime.now(), fail_chance=0.5):
    assert random.random() > fail_chance
    return "Databand checked at %s" % check_time



import os

from dbnd import output, pipeline, task
from targets import Target
from targets.types import PathStr


def save_my_custom_object(path, data):
    with open(path, "w") as fd:
        fd.writelines(data)


def load_my_custom_object(path):
    with open(path, "r") as fd:
        return fd.readlines()


@task
def read(loc):
    # type: (Target) -> str
    data = load_my_custom_object(loc.path)
    return " ".join(data)


@task
def write(data, res=output.data.require_local_access):
    save_my_custom_object(res.path, data)


@pipeline
def write_read(data=["abcd", "zxcvb"]):
    path = write(data)
    r = read(path.res)
    return r


@task
def read_dir(loc):
    # type: (Target) -> str
    data = []
    for f in os.listdir(loc.path):
        data += load_my_custom_object(os.path.join(loc.path, f))
    return " ".join(data)


@task
def write_dir(data, res=output.data.require_local_access[PathStr]):
    try:
        # Ensure dir exists
        os.mkdir(res)
    except Exception:
        pass
    for i in range(3):
        file_path = os.path.join(res, str(i) + ".data")
        save_my_custom_object(file_path, data)


@pipeline
def write_read_dir(data=["abcd", "zxcvb"]):
    dir = write_dir(data)
    r = read_dir(dir.res)
    return r



import typing

from dbnd._core.utils.basics.helpers import parse_bool


if typing.TYPE_CHECKING:
    from dbnd._core.configuration.dbnd_config import DbndConfig


class _DbndContextInitConfig(object):
    """
    intersects with CoreConfig, but this one can be loaded without being dependent on any other Task objects
    """

    def __init__(self, conf_provider):
        self._conf = c = conf_provider  # type: DbndConfig



import hashlib
import logging
import typing

from functools import partial
from itertools import chain

from dbnd._core.constants import RunState, TaskRunState
from dbnd._core.tracking.schemas.tracking_info_objects import (
    TargetInfo,
    TaskDefinitionInfo,
    TaskRunInfo,
    TaskRunParamInfo,
)
from dbnd._core.tracking.schemas.tracking_info_run import RunInfo
from dbnd._core.utils.string_utils import safe_short_string
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.traversing import traverse
from dbnd.api.tracking_api import InitRunArgs, TaskRunsInfo


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from typing import Dict, List
    from targets import Target
    from dbnd import Task
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


class TrackingInfoBuilder(object):
    def __init__(self, run):
        self.run = run  # type: DatabandRun

    def _run_to_run_info(self):
        # type: () -> RunInfo
        run = self.run
        task = run.driver_task_run.task
        context = run.context
        env = run.env
        return RunInfo(
            run_uid=run.run_uid,
            job_name=run.job_name,
            user=context.task_run_env.user,
            name=run.name,
            state=RunState.RUNNING,
            start_time=utcnow(),
            end_time=None,
            description=run.description,
            is_archived=run.is_archived,
            env_name=env.name,
            cloud_type=env.cloud_type,
            # deprecate and airflow
            dag_id=run.dag_id,
            execution_date=run.execution_date,
            cmd_name=context.name,
            driver_name=env.remote_engine or env.local_engine,
            # move to task
            target_date=task.task_target_date,
            version=task.task_version,
            # root and submitted by
            root_run=run.root_run_info,
            scheduled_run=run.scheduled_run_info,
            trigger="unknown",
            sends_heartbeat=run.sends_heartbeat,
            task_executor=run.task_executor_type,
        )

    def build_init_args(self):
        # type: () -> InitRunArgs

        run = self.run
        task_run_info = self.build_task_runs_info(run.task_runs)
        driver_task = run.driver_task_run.task
        init_args = InitRunArgs(
            run_uid=run.run_uid,
            root_run_uid=run.root_run_info.root_run_uid,
            task_runs_info=task_run_info,
            driver_task_uid=run.driver_task_run.task_run_uid,
            task_run_env=run.context.task_run_env,
            source=run.source,
            af_context=run.af_context,
        )

        if driver_task.is_submitter:
            init_args.new_run_info = self._run_to_run_info()

        if run.scheduled_run_info:
            init_args.scheduled_run_info = run.scheduled_run_info

        if run.root_run_info.root_task_run_uid:
            rel = (run.root_run_info.root_task_run_uid, init_args.driver_task_uid)
            task_run_info.parent_child_map.add(rel)
            task_run_info.upstreams_map.add(rel)

        return init_args

    def build_task_runs_info(self, task_runs, dynamic_task_run_update=False):
        # type: (List[TaskRun], bool) -> TaskRunsInfo
        run = self.run
        task_defs = {}
        all_task_models = {}
        all_targets = {}
        for task_run in task_runs:
            task = task_run.task
            # we process only tasks in current dag
            task_def_id = task.task_definition.full_task_family
            if task_def_id not in task_defs:
                task_defs[task_def_id] = task_to_task_def(run.context, task)

            self.task_to_targets(task, all_targets)
            all_task_models[task.task_id] = build_task_run_info(task_run)

        def _add_rel(rel_map, t_id_1, t_id_2):
            if t_id_1 in all_task_models or t_id_2 in all_task_models:
                tr_1 = run.get_task_run_by_id(t_id_1)
                tr_2 = run.get_task_run_by_id(t_id_2)
                if tr_1 and tr_2:
                    rel_map.add((tr_1.task_run_uid, tr_2.task_run_uid))

        # set children/upstreams maps
        upstreams_map = set()
        parent_child_map = set()

        for task_run in run.task_runs:
            task = task_run.task
            for t_id in task.task_meta.children:
                _add_rel(parent_child_map, task.task_id, t_id)

            task_dag = task.ctrl.task_dag
            for upstream in task_dag.upstream:
                _add_rel(upstreams_map, task.task_id, upstream.task_id)

        return TaskRunsInfo(
            run_uid=self.run.run_uid,
            root_run_uid=self.run.root_run_info.root_run_uid,
            task_run_env_uid=run.context.task_run_env.uid,
            task_definitions=[val for key, val in sorted(task_defs.items())],
            task_runs=[val for key, val in sorted(all_task_models.items())],
            targets=[val for key, val in sorted(all_targets.items())],
            parent_child_map=parent_child_map,
            upstreams_map=upstreams_map,
            dynamic_task_run_update=dynamic_task_run_update,
            af_context=run.af_context,
            parent_task_run_uid=run.root_run_info.root_task_run_uid,
            parent_task_run_attempt_uid=run.root_run_info.root_task_run_attempt_uid,
        )

    def task_to_targets(self, task, targets):
        # type: (Task, Dict[str, TargetInfo]) -> List[TargetInfo]
        """
        :param run:
        :param task:
        :param targets: all known targets for current run, so we have uniq list of targets (by path)
        :return:
        """

        run = self.run
        task_targets = []

        def process_target(target, name):
            # type: (Target, str) -> None
            target_path = str(target)
            dbnd_target = targets.get(target_path)
            if not dbnd_target:
                # we see this target for the first time
                target_task_run_uid = (
                    None  # let assume that Target is now owned by any task
                )
                # let try to find it's owner, so we create target that relates to some Task
                # if `task` is pipeline, the target owner is going to be different task
                if target.task:
                    target_task_run = run.get_task_run(target.task.task_id)
                    if target_task_run:
                        target_task_run_uid = target_task_run.task_run_uid

                dbnd_target = targets[target_path] = TargetInfo(
                    path=target_path,
                    created_date=utcnow(),
                    task_run_uid=target_task_run_uid,
                    parameter_name=name,
                )
                logger.debug(
                    "New Target: %s -> %s ->  %s",
                    target.task,
                    target_task_run_uid,
                    target_path,
                )
            task_targets.append(dbnd_target)

        rels = task.ctrl.relations
        for io_params in chain(rels.task_outputs.values(), rels.task_inputs.values()):
            for name, t in io_params.items():
                traverse(t, convert_f=partial(process_target, name=name))

        return task_targets


def task_to_task_def(ctx, task):
    # type: (DatabandContext, Task) -> TaskDefinitionInfo
    td = task.task_definition

    tracked_params = td.user_params if task.is_tracking_mode else td.all_task_params
    task_param_definitions = [value for key, value in sorted(tracked_params.items())]
    task_family = task.task_meta.task_family
    task_definition = TaskDefinitionInfo(
        task_definition_uid=td.task_definition_uid,
        class_version=task.task_class_version,
        family=task_family,
        module_source=td.task_module_code,
        module_source_hash=source_md5(td.task_module_code),
        name=task_family,
        source=td.task_source_code,
        source_hash=source_md5(td.task_source_code),
        type=task.task_meta.task_type,
        task_param_definitions=task_param_definitions,
    )
    return task_definition


def build_task_run_info(task_run):
    # type: (TaskRun) -> TaskRunInfo
    t = task_run.task
    tm = task_run.task.task_meta
    task_dag = t.ctrl.task_dag
    log_local, log_remote = task_run._get_log_files()

    task_params_values = dict(t._params.get_params_serialized())
    td = t.task_definition
    task_run_params = []
    tracked_params = td.user_params if t.is_tracking_mode else td.all_task_params
    for key, tdp in sorted(tracked_params.items()):
        param_meta = t._params.get_param_meta(tdp.name)
        if param_meta:
            value_source, value = param_meta.source, param_meta.value
        else:
            value_source, value = "", ""
        task_run_params.append(
            TaskRunParamInfo(
                parameter_name=tdp.name,
                value_origin=safe_short_string(str(value_source), max_value_len=5000),
                value=safe_short_string(
                    str(task_params_values.get(tdp.name, value))
                    if not tdp.hidden
                    else "***",
                    max_value_len=5000,
                ),
            )
        )

    return TaskRunInfo(
        run_uid=task_run.run.run_uid,
        task_definition_uid=task_run.task.task_definition.task_definition_uid,
        task_run_uid=task_run.task_run_uid,  # this is not the TaskRun uid
        task_run_attempt_uid=task_run.task_run_attempt_uid,  # this is not the TaskRun uid
        task_id=t.task_id,
        task_af_id=task_run.task_af_id,
        name=t.task_name,
        task_signature=tm.task_signature,
        task_signature_source=tm.task_signature_source,
        output_signature=tm.task_outputs_signature,
        command_line=tm.task_command_line,
        env=t.task_env.name,
        functional_call=tm.task_functional_call,
        has_downstreams=bool(task_dag.downstream),
        has_upstreams=bool(task_dag.upstream),
        state=TaskRunState.SCHEDULED
        if not task_run.is_reused
        else TaskRunState.SUCCESS,
        is_reused=task_run.is_reused,
        is_skipped=task_run.is_skipped,
        is_dynamic=task_run.is_dynamic,
        is_system=task_run.is_system,
        version=t.task_version,
        target_date=t.task_target_date,
        log_local=log_local,
        log_remote=log_remote,
        task_run_params=task_run_params,
        execution_date=task_run.run.execution_date,
        is_root=task_run.is_root,
    )


def source_md5(source_code):
    if source_code:
        try:
            return hashlib.md5(source_code.encode("utf-8")).hexdigest()
        except UnicodeDecodeError:
            return hashlib.md5(source_code).hexdigest()



import logging

from typing import TYPE_CHECKING, List

import six

from dbnd._core.configuration.dbnd_config import _ConfigLayer, config
from dbnd._core.errors import DatabandError
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterValue
from dbnd._core.task_build.task_signature import (
    Signature,
    build_signature,
    build_signature_from_values,
    user_friendly_signature,
)
from dbnd._core.utils.platform.windows_compatible.getuser import dbnd_getuser


if TYPE_CHECKING:
    from dbnd import Task
    from dbnd._core.task_build.task_definition import TaskDefinition
    from dbnd._core.task_build.task_params import TaskValueParams
    from dbnd._core.context.databand_context import DatabandContext

logger = logging.getLogger(__name__)

username = dbnd_getuser()


class TaskMeta(object):
    _current_task_creation_id = 0

    def __init__(
        self,
        task_name,
        task_family,
        task_definition,
        config_layer,
        task_params,
        task_config_override,
        dbnd_context,
        task_enabled=True,
        build_warnings=None,
        task_sections=None,
        task_call_source=None,
    ):
        super(TaskMeta, self).__init__()
        # we should not use it directly, the value in object can outdated
        self._task_params = task_params  # type: TaskValueParams
        self.class_task_params = list(
            task_params.class_params.values()
        )  # type: List[ParameterValue]

        self.config_layer = config_layer  # type: _ConfigLayer
        self.task_name = task_name  # type: str
        self.task_family = task_family  # type: str
        self.task_definition = task_definition  # type: TaskDefinition

        self.task_sections = task_sections
        self.task_config_override = task_config_override
        self.dbnd_context = dbnd_context  # type: DatabandContext

        self.children = set()
        self.build_warnings = build_warnings or []

        self.task_id = None
        self.task_signature = None
        self.task_signature_source = None

        self.task_outputs_signature = None
        self.task_outputs_signature_source = None

        self.task_enabled = task_enabled

        self.task_call_source = task_call_source

        sign_task_params = list(
            task_params.all_params.values()
        )  # type: List[ParameterValue]
        self.obj_key = self._calculate_task_meta_key(sign_task_params)
        # we want to have task id immediately
        self.initialize_task_id(
            [
                (p_value.name, p_value.parameter.signature(p_value.value))
                for p_value in sign_task_params
                if isinstance(p_value.parameter, ParameterDefinition)
                and p_value.parameter.significant
            ]
        )

        self.dag_id = self.task_name
        self.task_type = self.task_definition.task_class._conf__task_type_name

        self.task_command_line = None
        self.task_functional_call = None
        self.task_user = username

        # we count task meta creation
        # even if task_meta will not be used by TaskMetaClass when we already have created task
        # we will increase creation id
        # if t2 created after t1, t2.task_meta.task_creation_id > t1.task_meta.task_creation_id
        TaskMeta._current_task_creation_id += 1
        self.task_creation_id = TaskMeta._current_task_creation_id

    def add_child(self, task_id):
        self.children.add(task_id)

    def get_children(self):
        # type: (...)-> List[Task]
        tic = self.dbnd_context.task_instance_cache
        children = []
        for c_id in self.children:
            child_task = tic.get_task_by_id(c_id)
            if child_task is None:
                raise DatabandError(
                    "You have created %s in different dbnd_context, "
                    "can't find task object in current context!" % c_id
                )
            children.append(child_task)
        return children

    def _calculate_task_meta_key(self, sign_task_params):
        # type: (List[ParameterValue]) -> Signature
        params = [
            (p_value.name, p_value.parameter.signature(p_value.value))
            for p_value in sign_task_params
            if not p_value.parameter.is_output() and p_value.parameter.significant
        ]
        override_signature = {}
        for p_obj, p_val in six.iteritems(self.task_config_override):
            if isinstance(p_obj, ParameterDefinition):
                override_key = "%s.%s" % (p_obj.task_cls.get_task_family(), p_obj.name)
                override_value = (
                    p_val
                    if isinstance(p_val, six.string_types)
                    else p_obj.signature(p_val)
                )
            else:
                # very problematic approach till we fix the override structure
                override_key = str(p_obj)
                override_value = str(p_val)
            override_signature[override_key] = override_value

        params.append(("task_override", override_signature))

        # task schema id is unique per Class definition.
        # so if we have new implementation - we will not a problem with rerunning it
        full_task_name = "%s@%s(object=%s)" % (
            self.task_name,
            self.task_definition.full_task_family,
            str(id(self.task_definition)),
        )

        return build_signature(name=full_task_name, params=params)

    def initialize_task_id(self, params=None):
        name = self.task_name
        extra = {}
        if config.getboolean("task_build", "sign_with_full_qualified_name"):
            extra["full_task_family"] = self.task_definition.full_task_family
        if config.getboolean("task_build", "sign_with_task_code"):
            extra["task_code_hash"] = user_friendly_signature(
                self.task_definition.task_source_code
            )

        signature = build_signature(name=name, params=params, extra=extra)
        self.task_id, self.task_signature = (signature.id, signature.signature)

        self.task_signature_source = signature.signature_source

    def initialize_task_output_id(self, task_outputs):
        if task_outputs:
            sig = build_signature_from_values("task_outputs", task_outputs)
            self.task_outputs_signature = sig.signature
            self.task_outputs_signature_source = sig.signature_source
        else:
            self.task_outputs_signature = self.task_signature
            self.task_outputs_signature_source = self.task_signature_source

    def get_task_config_value(self, key):
        for section in self.task_sections:
            config_value = config.get_config_value(section, key)
            if config_value:
                return config_value
        return None



# -*- coding: utf-8 -*-
#
# Copyright 2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
The abstract :py:class:`Task` class.
It is a central concept of databand and represents the state of the workflow.
See :doc:`/tasks` for an overview.
"""
import logging
import typing

from typing import Dict, Optional

import six

from dbnd._core.constants import ParamValidation, TaskType
from dbnd._core.decorator.lazy_object_proxy import CallableLazyObjectProxy
from dbnd._core.decorator.task_decorator_spec import _TaskDecoratorSpec
from dbnd._core.errors import friendly_error
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.task_build.task_definition import TaskDefinition
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd._core.task_ctrl.task_auto_values import TaskAutoParamsReadWrite
from dbnd._core.task_ctrl.task_meta import TaskMeta
from dbnd._core.task_ctrl.task_parameters import TaskParameters
from dbnd._core.utils.basics.nothing import NOTHING


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.settings import DatabandSettings


@six.add_metaclass(TaskMetaclass)
class _BaseTask(object):
    # override to change get_task_family() -> changes task_family
    _conf__task_family = None
    _conf__track_source_code = True  # module/class/function user source code tracking
    task_namespace = NOTHING

    _conf__task_type_name = TaskType.python
    _conf__task_ui_color = None
    #####
    # override output path format
    _conf__base_output_path_fmt = None

    # stores call spec for the @task definition
    _conf__decorator_spec = None  # type: Optional[_TaskDecoratorSpec]

    _conf__tracked = True  # track task changes with TrackingStore
    _conf__no_child_params = False  # disable child scope params
    _conf_auto_read_params = True  # enables autoread of params.
    _conf_confirm_on_kill_msg = None  # get user confirmation on task kill if not empty
    _conf__require_run_dump_file = False

    # this is the state of autoread
    _task_auto_read_original = None
    ####
    # execution
    # will be used by Registry
    task_definition = None  # type: TaskDefinition
    is_tracking_mode = False  # type: bool

    # user can override this with his configuration
    defaults = None  # type: Dict[ParameterDefinition, any()]

    validate_no_extra_params = parameter.enum(ParamValidation).system(
        description="validate that all configured keys for a task have a matching parameter definition"
    )

    @classmethod
    def get_task_family(cls):
        return cls.task_definition.task_family

    @classmethod
    def get_full_task_family(cls):
        return cls.task_definition.full_task_family

    def __init__(self, **kwargs):
        super(_BaseTask, self).__init__()

        # most of the time we will use it as TaskMetaCtrl - we want this to be type hint!
        self.task_meta = kwargs["task_meta"]  # type: TaskMeta
        self._params = TaskParameters(self)

        for p_value in self.task_meta.class_task_params:
            setattr(self, p_value.name, p_value.value)

        self._task_auto_read_current = None
        self._task_auto_read_origin = None

    @property
    def task_id(self):
        return self.task_meta.task_id

    @property
    def task_signature(self):
        return self.task_meta.task_signature

    @property
    def task_name(self):
        return self.task_meta.task_name

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__
            and self._params.get_param_values() == other._params.get_param_values()
        )

    def __hash__(self):
        return hash(self.task_id)

    def __repr__(self):
        """
        Build a task representation like `MyTask(param1=1.5, param2='5')`
        """
        return "%s" % self.task_meta.task_id

    def __str__(self):
        """
        Build a task representation like `MyTask(param1=1.5, param2='5')`
        """
        return self.task_meta.task_id

    @property
    def friendly_task_name(self):
        if self.task_name != self.task_meta.task_family:
            return "%s[%s]" % (self.task_name, self.task_meta.task_family)
        return self.task_name

    def _auto_load_save_params(
        self, auto_read=False, save_on_change=False, normalize_on_change=False
    ):
        c = TaskAutoParamsReadWrite(
            task=self,
            auto_read=auto_read,
            save_on_change=save_on_change,
            normalize_on_change=normalize_on_change,
        )
        return c.auto_load_save_params()

    def clone(self, cls=None, output_params_to_clone=None, **kwargs):
        """
        Creates a new instance from an existing instance where some of the args have changed.

        There's at least two scenarios where this is useful (see test/clone_test.py):

        * remove a lot of boiler plate when you have recursive dependencies and lots of args
        * there's task inheritance and some logic is on the base class

        :param cls:
        :param kwargs:
        :return:
        """
        if cls is None:
            cls = self.__class__
        output_params_to_clone = output_params_to_clone or []
        new_k = {}
        for param_name, param_class in six.iteritems(
            cls.task_definition.all_task_params
        ):
            if param_class.is_output() and param_name not in output_params_to_clone:
                continue
            if param_name in kwargs:
                new_k[param_name] = kwargs[param_name]
            elif hasattr(self, param_name):
                new_k[param_name] = getattr(self, param_name)

        new_k["task_name"] = self.task_name
        return cls(**new_k)

    @property
    def settings(self):
        # type: () -> DatabandSettings

        return self.task_meta.dbnd_context.settings

    def __iter__(self):
        raise friendly_error.task_build.iteration_over_task(self)

    def _initialize(self):
        pass

    def _task_banner(self, banner, verbosity):
        """
        customize task banner
        """
        return

    def _validate(self):
        """
        will be called after after object is created
        :return:
        """
        return

    def simple_params_dict(self):
        return {p.name: p.value for p in self._params.task_meta.class_task_params}

    def load_task_runtime_values(self):
        for param, value in self._params.get_param_values():
            runtime_value = param.calc_runtime_value(value, task=self)
            setattr(self, param.name, runtime_value)



"""
Tracking store backends registry
"""

import logging
import typing

from dbnd._core.errors import friendly_error
from dbnd._core.plugin.dbnd_plugins import assert_web_enabled
from dbnd._core.tracking.backends import (
    CompositeTrackingStore,
    ConsoleStore,
    FileTrackingStore,
    TbSummaryFileStore,
    TrackingStore,
    TrackingStoreThroughChannel,
)


logger = logging.getLogger(__name__)


_BACKENDS_REGISTRY = {
    "file": FileTrackingStore,
    "console": ConsoleStore,
    "debug": TrackingStoreThroughChannel.build_with_console_debug_channel,
    ("api", "web"): TrackingStoreThroughChannel.build_with_web_channel,
    ("api", "proto"): TrackingStoreThroughChannel.build_with_proto_web_channel,
    ("api", "disabled"): TrackingStoreThroughChannel.build_with_disabled_channel,
}


def register_store(name, store_builder):
    if name in _BACKENDS_REGISTRY:
        raise Exception("Already registered")
    _BACKENDS_REGISTRY[name] = store_builder


def get_tracking_store(
    tracking_store_names, api_channel_name, tracker_raise_on_error, remove_failed_store
):
    # type: (...) -> TrackingStore
    tracking_store_instances = []
    for name in tracking_store_names:
        if name == "api":
            name = (name, api_channel_name)
        if name == ("api", "db"):
            assert_web_enabled(
                "It is required when trying to use local db connection (tracker_api=db)."
            )
        tracking_store_builder = _BACKENDS_REGISTRY.get(name)
        if tracking_store_builder is None:
            friendly_error.config.wrong_store_name(name)
        instance = tracking_store_builder()
        tracking_store_instances.append(instance)

    # if tracker_raise_on_error is False - must use CompositeTrackingStore
    # regardless of tracking_stores - only CompositeTrackingStore suppresses errors
    return (
        tracking_store_instances[0]
        if len(tracking_store_instances) == 1 and tracker_raise_on_error
        else CompositeTrackingStore(
            tracking_stores=tracking_store_instances,
            raise_on_error=tracker_raise_on_error,
            remove_failed_store=remove_failed_store,
        )
    )



import datetime
import logging
import random
import typing
import warnings

from typing import Dict

from dbnd._core.constants import OutputMode, _TaskParamContainer
from dbnd._core.current import get_databand_run
from dbnd._core.errors.friendly_error.task_build import incomplete_output_found_for_task
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.parameter.parameter_builder import output, parameter
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    ParameterScope,
)
from dbnd._core.settings.env import EnvConfig
from dbnd._core.task.base_task import _BaseTask
from dbnd._core.task_ctrl.task_ctrl import TaskCtrl
from dbnd._core.task_ctrl.task_output_builder import calculate_path
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.traversing import flatten
from targets import target
from targets.target_config import TargetConfig, folder
from targets.value_meta import ValueMetaConf
from targets.values.version_value import VersionStr


if typing.TYPE_CHECKING:
    from dbnd._core.task_ctrl.task_dag import _TaskDagNode
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.decorator.safe_task_call import TaskCallState

DEFAULT_CLASS_VERSION = ""

logger = logging.getLogger(__name__)


class Task(_BaseTask, _TaskParamContainer):
    """
    This is the base class of all dbnd Tasks, the base unit of work in databand.

    A dbnd Task describes a unit or work.

    The key methods of a Task, which must be implemented in a subclass are:

    * :py:meth:`run` - the computation done by this task.

    Each :py:class:`~dbnd.parameter` of the Task should be declared as members:

    .. code:: python

        class MyTask(dbnd.Task):
            count = dbnd.parameter[int]
            second_param = dbnd.parameter[str]

    In addition to any declared properties and methods, there are a few
    non-declared properties, which are created by the :py:class:`TaskMetaclass`
    metaclass:

    """

    """
        This value can be overriden to set the namespace that will be used.
        (See :ref:`Task.namespaces_famlies_and_ids`)
        If it's not specified and you try to read this value anyway, it will return
        garbage. Please use :py:meth:`get_task_namespace` to read the namespace.

        Note that setting this value with ``@property`` will not work, because this
        is a class level value.
    """

    _task_band_result = output(default=None, system=True)
    _meta_output = output(
        system=True,
        output_name="meta",
        output_ext="",
        target_config=folder,
        significant=False,
        description="Location of all internal outputs (e.g. metrics)",
    )
    task_band = output.json(output_name="band", system=True)

    task_enabled = parameter.system(scope=ParameterScope.children)[bool]
    task_enabled_in_prod = parameter.system(scope=ParameterScope.children)[bool]

    # for permanent bump of task version use Task.task_class_version
    task_version = parameter(
        description="task version, directly affects task signature ",
        scope=ParameterScope.children,
    )[VersionStr]

    task_class_version = parameter.value(
        default=DEFAULT_CLASS_VERSION,
        system=True,
        description="task code version, "
        "use while you want persistent change in your task version",
    )

    task_env = parameter.value(
        description="task environment name", scope=ParameterScope.children
    )[EnvConfig]

    task_target_date = parameter(
        description="task data target date", scope=ParameterScope.children
    )[datetime.date]

    task_airflow_op_kwargs = parameter.system(
        default=None, description="airflow operator kwargs"
    )[Dict[str, object]]

    task_config = parameter.system(empty_default=True)[Dict]
    task_is_system = parameter.system(default=False)[bool]

    task_in_memory_outputs = parameter.system(
        scope=ParameterScope.children, description="Store all task outputs in memory"
    )[bool]
    task_is_dynamic = parameter.system(
        scope=ParameterScope.children,
        description="task was executed from within another task",
    )[bool]

    # for example: if task.run doesn't have access to databand, we can't run runtime tasks
    task_supports_dynamic_tasks = parameter.system(
        default=True, description="indicates if task can run dynamic databand tasks"
    )[bool]

    task_retries = parameter.system(
        description="Total number of attempts to run the task. So task_retries=3 -> task can fail 3 times before we give up"
    )[int]

    task_retry_delay = parameter.system(
        description="timedelta to wait before retrying a task. Example: 5s"
    )[datetime.timedelta]

    _dbnd_call_state = None  # type: TaskCallState

    def __init__(self, **kwargs):
        super(Task, self).__init__(**kwargs)
        self.ctrl = TaskCtrl(self)

    def band(self):
        """
        Please, do not override this function only in Pipeline/External tasks! we do all wiring work in Meta classes only
        Our implementation should never be coupled to code!
        :return:
        """
        return

    def run(self):
        """
        The task run method, to be overridden in a subclass.

        See :ref:`Task.run`
        """
        pass  # default impl

    @property
    def task_outputs(self):
        """
        The output that this Task produces.

        The output of the Task determines if the Task needs to be run--the task
        is considered finished iff the outputs all exist.
        See :ref:`Task.task_outputs`
        """
        return self.ctrl.relations.task_outputs_user

    @property
    def task_dag(self):
        # type: (...)->_TaskDagNode
        return self.ctrl.task_dag

    def _complete(self):
        """
        If the task has any outputs, return ``True`` if all outputs exist.
        Otherwise, return ``False``.

        However, you may freely override this method with custom logic.
        """
        # we check only user side task outputs
        # all system tasks outputs are not important (if the exists or not)
        # user don't see them
        outputs = [
            o for o in flatten(self.task_outputs) if not o.config.overwrite_target
        ]
        if len(outputs) == 0:
            if not self.task_band:
                warnings.warn(
                    "Task %r without outputs has no custom complete() and no task band!"
                    % self,
                    stacklevel=2,
                )
                return False
            else:
                return self.task_band.exists()

        incomplete_outputs = [str(o) for o in outputs if not o.exists()]

        num_of_incomplete_outputs = len(incomplete_outputs)

        if 0 < num_of_incomplete_outputs < len(outputs):
            complete_outputs = [str(o) for o in outputs if o.exists()]
            exc = incomplete_output_found_for_task(
                self.task_meta.task_name, complete_outputs, incomplete_outputs
            )

            if self.task_env.settings.run.validate_task_outputs_on_build:
                raise exc
            else:
                logger.warning(str(exc))

        return num_of_incomplete_outputs == 0

    @property
    def current_task_run(self):
        # type: ()->TaskRun
        return get_databand_run().get_task_run(self.task_id)

    def _output(self):
        """
        The default output that this Task produces. Use outputs! Override only if you are writing "base" class
        """
        return NOTHING

    def _requires(self):
        """
        Override in "template" tasks which themselves are supposed to be
        subclassed

        Must return an iterable which among others contains the _requires() of
        the superclass.
        See :ref:`Task.requires`
        """
        pass

    def _task_submit(self):
        """
        Task submission logic, by default we just call -> _task_run() -> run()
        """
        return self._task_run()

    def _task_run(self):
        # bring all relevant files
        self.current_task_run.sync_local.sync_pre_execute()
        with self._auto_load_save_params(
            auto_read=self._conf_auto_read_params, save_on_change=True
        ):
            result = self.run()

        self.current_task_run.sync_local.sync_post_execute()
        # publish all relevant files
        return result

    def set_upstream(self, task_or_task_list):
        self.task_dag.set_upstream(task_or_task_list)

    def set_downstream(self, task_or_task_list):
        self.task_dag.set_downstream(task_or_task_list)

    def __lshift__(self, other):
        return self.set_upstream(other)

    def __rshift__(self, other):
        return self.set_downstream(other)

    def set_global_upstream(self, task_or_task_list):
        self.task_dag.set_global_upstream(task_or_task_list)

    @property
    def metrics(self):
        # backward compatible code
        return self.current_task_run.tracker

    def log_dataframe(
        self,
        key,
        df,
        with_preview=True,
        with_schema=True,
        with_size=True,
        with_stats=False,
    ):
        meta_conf = ValueMetaConf(
            log_preview=with_preview,
            log_schema=with_schema,
            log_size=with_size,
            log_stats=with_stats,
        )
        self.metrics.log_data(key, df, meta_conf=meta_conf)

    def log_metric(self, key, value, source=None):
        """
        Logs the passed-in parameter under the current run, creating a run if necessary.
        :param key: Parameter name (string)
        :param value: Parameter value (string)
        """
        return self.metrics.log_metric(key, value, source=source)

    def log_system_metric(self, key, value):
        """Shortcut for log_metric(..., source="system") """
        return self.log_metric(key, value, source="system")

    def log_artifact(self, name, artifact):
        """Log a local file or directory as an artifact of the currently active run."""
        return self.metrics.log_artifact(name, artifact)

    def get_template_vars(self):
        # TODO: move to cached version, (after relations are built)
        base = {
            "task": self,
            "task_family": self.task_meta.task_family,
            "task_name": self.task_meta.task_name,
            "task_signature": self.task_meta.task_signature,
            "task_id": self.task_meta.task_id,
        }
        base.update(self._params.get_params_serialized(input_only=True))
        if self.task_target_date is None:
            base["task_target_date"] = "input"
        return base

    def on_kill(self):
        """
        Override this method to cleanup subprocesses when a task instance
        gets killed. Any use of the threading, subprocess or multiprocessing
        module within an operator needs to be cleaned up or it will leave
        ghost processes behind.
        """
        pass

    def _get_task_output_path_format(self, output_mode):
        if self.task_env.production and output_mode == OutputMode.prod_immutable:
            return self.settings.output.path_prod_immutable_task
        return self._conf__base_output_path_fmt or self.settings.output.path_task

    def get_target(self, name, config=None, output_ext=None, output_mode=None):
        name = name or "tmp/dbnd-tmp-%09d" % random.randint(0, 999999999)
        config = config or TargetConfig()
        path_pattern = self._get_task_output_path_format(output_mode)

        path = calculate_path(
            task=self,
            name=name,
            output_ext=output_ext,
            is_dir=config.folder,
            path_pattern=path_pattern,
        )

        return target(path, config=config)

    def get_root(self):
        return self.task_env.root

    def _initialize(self):
        super(Task, self)._initialize()
        self.ctrl._initialize_task()

    def _should_run(self):
        if not self.task_enabled:
            return False

        if self.task_env.production:
            return self.task_enabled_in_prod or self.settings.run.enable_prod

        return True

    @dbnd_handle_errors(exit_on_error=False)
    def dbnd_run(self):
        # type: (...)-> DatabandRun
        """
        Run task via Databand execution system
        """
        # this code should be executed under context!
        from dbnd._core.current import get_databand_context

        ctx = get_databand_context()
        result = ctx.dbnd_run_task(self)
        return result


Task.task_definition.hidden = True

TASK_PARAMS_COUNT = len(
    [
        v
        for k, v in Task.__dict__.items()
        if isinstance(v, ParameterDefinition) and v.value_type_str != "Target"
    ]
)



import logging
import subprocess
import sys

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.context import bootstrap
from dbnd._core.errors import (
    DatabandConfigError,
    DatabandError,
    DatabandRunError,
    DatabandRuntimeError,
    DatabandSystemError,
    friendly_error,
    get_help_msg,
    show_exc_info,
)
from dbnd._core.errors.base import DatabandRunError
from dbnd._core.errors.errors_utils import (
    get_user_frame_info_str,
    nested_exceptions_str,
)
from dbnd._core.utils import console_utils, seven
from dbnd._core.utils.basics.format_exception import format_exception_as_str
from dbnd._core.utils.basics.helpers import indent


logger = logging.getLogger(__name__)


def get_databand_error_mesage(ex, args=None, sys_exit=True):
    args = args or sys.argv
    please_report = False
    print_source = True

    if isinstance(ex, DatabandRunError):
        # we already printed all information!
        return (
            "There is an error! Your run has failed!",
            DatabandExitCodes.execution_failed,
        )

    if isinstance(ex, DatabandRuntimeError):
        code = DatabandExitCodes.execution_failed
    elif isinstance(ex, DatabandConfigError):
        code = DatabandExitCodes.configuration_error
    elif isinstance(ex, DatabandRunError):
        code = DatabandExitCodes.execution_failed
    elif isinstance(ex, DatabandSystemError):
        code = DatabandExitCodes.error
        please_report = True
    elif isinstance(ex, DatabandError):
        code = DatabandExitCodes.error
    elif ex.__class__.__name__ == "NoCredentialsError":  # aws
        code = DatabandExitCodes.configuration_error
        ex = friendly_error.config.no_credentials()
        print_source = False
    else:
        please_report = True
        code = DatabandExitCodes.unknown_error

    msg = str(ex)

    extra_msg_lines = []

    nested_exceptions = nested_exceptions_str(ex)
    if nested_exceptions:
        extra_msg_lines.append("Caused by: \n%s\n" % indent(nested_exceptions, "\t"))

    help_msg = get_help_msg(ex)
    if help_msg:
        extra_msg_lines.append(" Help: \n%s\n" % indent(help_msg, "\t"))

    user_frame_info_str = get_user_frame_info_str(ex)
    if user_frame_info_str and print_source:
        extra_msg_lines.append("Source: \n%s\n" % indent(user_frame_info_str, "\t"))

    # if we crashed before finishing bootstrap we probably want to see the full trace, and we could have failed during config init so the verbose flag does nothing
    if (
        show_exc_info(ex)
        or config.getboolean("databand", "verbose")
        or not bootstrap._dbnd_bootstrap
    ):
        error_info = sys.exc_info()
        extra_msg_lines.append(format_exception_as_str(error_info))

    if please_report:
        extra_msg_lines.append(
            " Please report it to support@databand.ai or appropriate slack channel!"
        )
    msg = (
        "There is an error! Your run has failed with {exc_type}\n"
        "{sep}\n"
        " Command line: {command_line}\n"
        " Failure:\n{msg}\n\n"
        "{extra_msg}\n"
        "{sep}\n"
        "".format(
            sep=console_utils.ERROR_SEPARATOR,
            command_line=subprocess.list2cmdline(args or []),
            sep_small=console_utils.ERROR_SEPARATOR_SMALL,
            msg=console_utils.bold(indent(msg, "\t")),
            exc_type=ex.__class__.__name__,
            extra_msg="\n ".join(extra_msg_lines),
        )
    )
    return msg, code


@seven.contextlib.contextmanager
def dbnd_handle_errors(exit_on_error=True):
    try:
        yield
    except Exception as ex:
        if not hasattr(ex, "_dbnd_error_handled"):
            msg, code = get_databand_error_mesage(ex=ex, args=None)
            logger.error(msg)
            setattr(ex, "_dbnd_error_handled", True)
        if exit_on_error:
            sys.exit(1)
        raise


class DatabandExitCodes(object):
    unknown_error = 4
    error = 5
    databand_system = 6
    task_runtime = 7
    configuration_error = 8
    execution_failed = 10



import logging
import os
import typing

from contextlib import contextmanager

import six

from dbnd._core.log.logging_utils import find_handler, redirect_stderr, redirect_stdout
from dbnd._core.settings import LocalEnvConfig
from dbnd._core.settings.log import _safe_is_typeof
from dbnd._core.task_run.log_preview import read_dbnd_log_preview
from dbnd._core.task_run.task_run_ctrl import TaskRunCtrl
from dbnd._core.utils.string_utils import merge_dbnd_and_spark_logs
from targets import target


if typing.TYPE_CHECKING:
    pass
logger = logging.getLogger(__name__)

CURRENT_TASK_HANDLER_LOG = None


class TaskRunLogManager(TaskRunCtrl):
    def __init__(self, task_run):
        super(TaskRunLogManager, self).__init__(task_run)

        self.local_log_file = self.task_run.local_task_run_root.partition(
            name="%s.log" % task_run.attempt_number
        )

        if os.getenv("DBND__LOG_SPARK"):
            self.local_spark_log_file = self.task_run.local_task_run_root.partition(
                name="%s-spark.log" % task_run.attempt_number
            )

        self.local_heartbeat_log_file = self.task_run.local_task_run_root.partition(
            name="%s.heartbeat.log" % task_run.attempt_number
        )
        self.remote_log_file = None
        if not isinstance(self.task.task_env, LocalEnvConfig):
            self.remote_log_file = self.task_run.attempt_folder.partition(
                "%s.log" % task_run.attempt_number
            )

        # file handler for task log
        # if set -> we are in the context of capturing
        self._log_task_run_into_file_active = False

    @contextmanager
    def capture_stderr_stdout(self, logging_target=None):
        #  redirecting all messages from sys.stderr/sys.stdout into logging_target
        # WARNING: this can create stdout/stderr loop if :
        #     we redirect into logger, and logger is writing to current sys.stdout that is current redirect
        #     ( not the original one before the wrapping)
        #    -> LOOP:   some print ->  redirect_stdout() -> logger -> print -> redirect_stdout() ...
        if not self.task.settings.log.capture_stdout_stderr:
            yield None
            return

        airflow_root_console_handler = find_handler(logging.root, "console")
        if _safe_is_typeof(airflow_root_console_handler, "RedirectStdHandler"):
            yield None
            return

        logging_target_stdout = logging_target or logging.getLogger("dbnd.stdout")
        logging_target_stderr = logging_target or logging.getLogger("dbnd.stderr")
        with redirect_stdout(logging_target_stderr, logging.INFO), redirect_stderr(
            logging_target_stdout, logging.WARN
        ):
            yield

    @contextmanager
    def capture_task_log(self):
        global CURRENT_TASK_HANDLER_LOG
        log_file = self.local_log_file

        log_settings = self.task.settings.log
        if (
            self._log_task_run_into_file_active
            or not log_settings.capture_task_run_log
            or not log_file
        ):
            yield None
            return

        if (
            # there is no really any good way to find if this is runtime tracking for airflow operator
            # should be refactored as soon as possible
            self.task.get_task_family().endswith("_execute")
        ):
            yield None
            return

        handler = log_settings.get_task_log_file_handler(log_file)
        if not handler:
            yield None
            return
        target_logger = logging.root
        logger.debug("Capturing task log into '%s'", log_file)

        try:
            self.attach_spark_logger()
            target_logger.addHandler(handler)
            self._log_task_run_into_file_active = True
            CURRENT_TASK_HANDLER_LOG = handler

            with self.capture_stderr_stdout():
                yield handler
        except Exception as task_ex:
            CURRENT_TASK_HANDLER_LOG = None
            raise task_ex
        finally:
            self.detach_spark_logger()
            try:
                target_logger.removeHandler(handler)
                handler.close()
            except Exception:
                logger.error("Failed to close file handler for log %s", log_file)
            self._log_task_run_into_file_active = False
            self._upload_task_log_preview()

    def attach_spark_logger(self):
        if os.getenv("DBND__LOG_SPARK"):
            spark_log_file = self.local_spark_log_file
            try:
                from pyspark.sql import SparkSession

                spark = SparkSession.builder.getOrCreate()
                log4j = spark._jvm.org.apache.log4j

                spark_logger = log4j.Logger.getLogger("org.apache.spark")

                pattern = "[%d] {%c,%C{1}} %p - %m%n"
                file_appender = log4j.FileAppender()

                file_appender.setFile(spark_log_file.path)
                file_appender.setName(spark_log_file.path)
                file_appender.setLayout(log4j.PatternLayout(pattern))
                file_appender.setThreshold(log4j.Priority.toPriority("INFO"))
                file_appender.activateOptions()
                spark_logger.addAppender(file_appender)
            except Exception as task_ex:
                logger.error(
                    "Failed to attach spark logger for log %s: %s",
                    spark_log_file,
                    task_ex,
                )

    def detach_spark_logger(self):
        if os.getenv("DBND__LOG_SPARK"):
            spark_log_file = self.local_spark_log_file
            try:
                from pyspark.sql import SparkSession

                spark = SparkSession.builder.getOrCreate()

                jvm = spark._jvm
                log4j = jvm.org.apache.log4j
                spark_logger = log4j.Logger.getLogger("org.apache.spark")

                spark_logger.removeAppender(spark_log_file.path)
            except Exception as task_ex:
                logger.error(
                    "Failed to detach spark logger for log %s: %s",
                    spark_log_file,
                    task_ex,
                )

    def _upload_task_log_preview(self):
        try:
            log_body = self.read_log_body()
            self.write_remote_log(log_body)
            self.save_log_preview(log_body)
        except Exception as save_log_ex:
            logger.error("failed to save log preview for %s:%s", self, save_log_ex)

    def read_log_body(self):
        try:
            spark_log_file = (
                self.local_spark_log_file.path if os.getenv("DBND__LOG_SPARK") else None
            )
            return read_dbnd_log_preview(self.local_log_file.path, spark_log_file)

        except Exception as ex:
            logger.error(
                "Failed to read log (%s) for %s: %s",
                self.local_log_file.path,
                self.task,
                ex,
            )
            return None

    def write_remote_log(self, log_body):
        if self.task.settings.log.remote_logging_disabled or not self.remote_log_file:
            return

        try:
            self.remote_log_file.write(log_body)
        except Exception as ex:
            # todo add remote log path to error
            logger.warning("Failed to write remote log for %s: %s", self.task, ex)

    def save_log_preview(self, log_body):
        self.task_run.tracker.save_task_run_log(log_body, self.local_log_file)



import logging

from dbnd._core.tracking.backends import TrackingStore


logger = logging.getLogger(__name__)


class TbSummaryFileStore(TrackingStore):
    def init_run(self, run):
        from dbnd._core.tools.tensorboard import save_task_graph

        save_task_graph(
            run.task,
            comment=run.run_id,
            log_dir=run.local_driver_root.folder("tensorboard.summary"),
        )



import logging
import os
import sys

from contextlib import contextmanager

from dbnd._core.utils.platform import windows_compatible_mode
from targets.utils.path import safe_mkdirs


logger = logging.getLogger(__name__)


def _reset_handlers():
    root = logging.root
    map(root.removeHandler, root.handlers[:])
    map(root.removeFilter, root.filters[:])


def symlink_latest_log(log_file, latest_log=None):
    if windows_compatible_mode:
        # there is no symlinks in windows
        return

    if latest_log is None:
        log_directory, log_filename = os.path.split(log_file)
        latest_log = os.path.join(log_directory, "latest")

    try:
        # if symlink exists but is stale, update it
        if os.path.islink(latest_log):
            if os.readlink(latest_log) != log_file:
                os.unlink(latest_log)
                os.symlink(log_file, latest_log)
        elif os.path.isdir(latest_log) or os.path.isfile(latest_log):
            logging.warning(
                "%s already exists as a dir/file. Skip creating symlink.", latest_log
            )
        else:
            os.symlink(log_file, latest_log)
    except OSError:
        logging.warning(
            "OSError while attempting to symlink " "the latest %s" % latest_log
        )


def setup_logs_dir(log_dir):
    if not os.path.exists(log_dir):
        safe_mkdirs(log_dir, 0o777)


def setup_log_file(log_file):
    setup_logs_dir(os.path.dirname(log_file))


def create_file_handler(log_file, fmt=None):
    fmt = fmt or "%(asctime)s %(levelname)s - %(message)s"
    formatter = logging.Formatter(fmt=fmt)

    # "formatter": log_settings.file_formatter,
    log_file = str(log_file)
    setup_log_file(log_file)
    handler = logging.FileHandler(filename=log_file, encoding="utf-8")
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    return handler


@contextmanager
def override_log_formatting(log_format):
    original_formatters = [handler.formatter for handler in logger.root.handlers]
    try:
        raw_formatter = logging.Formatter(log_format)
        for handler in logger.root.handlers:
            handler.setFormatter(raw_formatter)

        yield
    finally:
        for handler, formatter in zip(logger.root.handlers, original_formatters):
            handler.setFormatter(formatter)


def raw_log_formatting():
    return override_log_formatting("%(message)s")


class StreamLogWriter(object):
    encoding = False

    """
    Allows to redirect stdout and stderr to logger
    """

    def __init__(self, logger, level):
        """
        :param log: The log level method to write to, ie. log.debug, log.warning
        :return:
        """
        self.logger = logger
        self.level = level
        self._buffer = str()
        self.skip_next_msg = 0

    def write(self, message):
        """
        Do whatever it takes to actually log the specified logging record
        :param message: message to log
        """
        if not message.endswith("\n"):
            self._buffer += message
            return

        self._buffer += message
        log_msg = self._buffer.rstrip()

        # we want to prevent  stderr -> logger -> FAILURE with stderr print (inside logging.py) -> stderr -> recursion
        # so the moment we understand that there is and logger error -> we stop redirecting for the next 10 messages
        if log_msg == "--- Logging error ---":
            self.skip_next_msg = 100
            sys.__stderr__.write(
                "Logger have an error, disable stream redirect for next 100 lines\n"
            )

        if not self.skip_next_msg:
            self.logger.log(self.level, self._buffer.rstrip())
        else:
            self.skip_next_msg -= 1
            sys.__stderr__.write("%s\n" % log_msg)
        self._buffer = str()

    def flush(self):
        """
        Ensure all logging output has been flushed
        """
        if len(self._buffer) > 0:
            self.logger.log(self.level, self._buffer)
            self._buffer = str()

    def isatty(self):
        """
        Returns False to indicate the fd is not connected to a tty(-like) device.
        For compatibility reasons.
        """
        return False


@contextmanager
def redirect_stdout(logger, level):
    writer = StreamLogWriter(logger, level)
    original = sys.stdout
    try:
        sys.stdout = writer
        yield
    finally:
        sys.stdout = original


@contextmanager
def redirect_stderr(logger, level):
    writer = StreamLogWriter(logger, level)
    original = sys.stderr
    try:
        sys.stderr = writer
        yield
    finally:
        sys.stderr = original


def set_module_logging_to_debug(modules):
    for m in modules:
        logging.getLogger(m.__name__).setLevel(logging.DEBUG)


class TaskContextFilter(logging.Filter):
    """
    adding 'task' variable to every record
    """

    task = "main"

    def filter(self, record):
        if self.task is not None:
            record.task = self.task
        return True

    @classmethod
    @contextmanager
    def task_context(cls, task_id):
        original_task = cls.task
        cls.task = task_id
        try:
            yield cls
        finally:
            cls.task = original_task


def find_handler(logger, handler_name):
    for h in logger.handlers:
        if h.name == handler_name:
            return h
    return None


SENTRY_TOUCHED = False


def try_init_sentry():
    """
    Initiate the sentry sdk to track the errors in the system.
    This function is idempotent - we allowed running it once in a single run.
    """
    global SENTRY_TOUCHED
    if SENTRY_TOUCHED:
        return

    from dbnd._core.settings import LoggingConfig

    logging_config = LoggingConfig.current()

    sentry_url = logging_config.sentry_url

    if sentry_url:
        try:
            import sentry_sdk
        except ImportError:
            logger.warning(
                "sentry_sdk is not installed. "
                "Trying to init Sentry because the LoggingConfig.sentry_url configuration is set. "
                "Help: To use Sentry please run `pip install sentry_sdk`."
            )

        else:
            # This is the function that initiate sentry with the config
            # Add debug=True if you want to debug sentry
            sentry_sdk.init(
                dsn=sentry_url,
                environment=logging_config.sentry_env,
                debug=logging_config.sentry_debug,
            )

            # sentry doesn't work good with coloring - clearing the logs colors
            logging_config.disable_colors = True

            logger.debug(
                "running with sentry. dsn={dsn}, environment={environment}".format(
                    dsn=sentry_url, environment=logging_config.sentry_env
                )
            )

    SENTRY_TOUCHED = True



def getrlimit(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()


def getrusage(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()


def setrlimit(*args, **kwargs):  # real signature unknown
    raise NotImplementedError()



import functools
import logging
import os
import re

from collections import Mapping

import attr
import six

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.configuration.config_store import _ConfigStore
from dbnd._core.configuration.config_value import ConfigValue
from dbnd._core.configuration.environ_config import (
    get_dbnd_custom_config,
    get_dbnd_environ_config_file,
    in_quiet_mode,
    is_unit_test_mode,
)
from dbnd._core.errors import DatabandConfigError
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.utils.project.project_fs import (
    databand_config_path,
    databand_system_path,
)
from dbnd._vendor.snippets.airflow_configuration import expand_env_var
from targets import LocalFileSystem, target
from targets.pipes import Text


logger = logging.getLogger(__name__)
_DBND_ENVIRON_RE = re.compile(r"^DBND__([\w\d_\-]+)__([\w\d_\-]+)$")
_SECTION_NAME_RE = re.compile(r"(.*)\.([^.]*)")


def _default_configuration_paths():
    # we always have "library config"
    yield databand_config_path("databand-core.cfg")

    system_config = os.path.expanduser("/etc/databand.cfg")
    if os.path.isfile(system_config):
        yield system_config

    # now we can start to look for project configs
    dbnd_project_config = get_dbnd_project_config()

    possible_locations = [
        databand_system_path("databand-system.cfg"),
        dbnd_project_config.dbnd_project_path("conf", "databand.cfg"),  # deprecated
        dbnd_project_config.dbnd_project_path("databand.cfg"),  # deprecated
        get_dbnd_custom_config(),
        dbnd_project_config.dbnd_project_path("project.cfg"),
    ]
    env_config = get_dbnd_environ_config_file()
    if env_config:
        possible_locations.append(env_config)
    for value in possible_locations:
        value = expand_env_var(value)
        if os.path.isfile(value):
            yield value

    user_config = expand_env_var("~/.dbnd/databand.cfg")
    if os.path.isfile(user_config):
        yield user_config

    if is_unit_test_mode():
        tests_config_path = databand_system_path("databand-test.cfg")
        if os.path.exists(tests_config_path):
            yield tests_config_path


def read_from_config_stream(config_fp, source="<stream>"):
    """
    Read config from config file (.ini, .cfg)
    """
    from backports.configparser import ConfigParser

    parser = ConfigParser()
    parser._read(config_fp, source)

    source = "config[{file_name}]".format(file_name=os.path.basename(str(source)))
    new_config = _ConfigStore()
    for section in parser.sections():
        for option in parser.options(section):
            value = parser.get(section, option)
            new_config.set_config_value(
                section, option, ConfigValue(value, source, require_parse=True)
            )

    return new_config


def read_from_config_file(config_file_target):
    """
    Read config from config file (.ini, .cfg)
    """

    # Check if config files are inside zip first - can happen if we run in fat wheel
    if (
        isinstance(config_file_target.fs, LocalFileSystem)
        and ".zip/" in config_file_target.path
    ):
        zip_file_path, config_path_inside_zip = config_file_target.path.split(".zip/")
        if os.path.exists(zip_file_path + ".zip"):
            import zipfile

            archive = zipfile.ZipFile(zip_file_path + ".zip", "r")
            if config_path_inside_zip not in archive.namelist():
                raise DatabandConfigError(
                    "Failed to read configuration file at %s, file not found!"
                    % config_file_target
                )
            archive = zipfile.ZipFile(zip_file_path + ".zip", "r")
            with archive.open(config_path_inside_zip) as file_io:
                return read_from_config_stream(
                    Text.pipe_reader(file_io), str(config_file_target)
                )

    if not config_file_target.exists():
        raise DatabandConfigError(
            "Failed to read configuration file at %s, file not found!"
            % config_file_target
        )
    with config_file_target.open("r") as fp:
        return read_from_config_stream(fp, str(config_file_target))


def read_from_config_files(config_files):
    files_to_load = [target(f) for f in config_files]
    configs = []

    if not in_quiet_mode():
        logger.info(
            "Reading configuration from: \n\t%s\n", "\n\t".join(map(str, files_to_load))
        )

    for f in files_to_load:
        try:
            configs.append(read_from_config_file(f))
        except DatabandConfigError:
            raise
        except Exception as ex:
            raise DatabandConfigError(
                "Failed to read configuration file at %s: %s" % (f, ex),
                nested_exceptions=ex,
            )

    merged_file_config = functools.reduce((lambda x, y: x.update(y)), configs)
    return merged_file_config


def read_environ_config():
    """
    Read configuration from process environment
    Every env var in following format will be added to config
    $DBND__SECTION__KEY=value  (please notice double underscore "__")
    :return:
    """
    return get_environ_config_from_dict(os.environ, "environ")


def get_environ_config_from_dict(env_dict, source_prefix):
    dbnd_environ = _ConfigStore()
    for key, value in six.iteritems(env_dict):
        if not key.startswith("DBND__"):
            continue
        # must have format DBND__{SECTION}__{KEY} (note double underscore)
        dbnd_key_var = _DBND_ENVIRON_RE.match(key)
        if dbnd_key_var:
            section, key = dbnd_key_var.group(1), dbnd_key_var.group(2)
            dbnd_environ.set_config_value(
                section,
                key,
                ConfigValue(
                    value,
                    source="{}[{}]".format(source_prefix, key),
                    require_parse=True,
                ),
            )
        else:
            # check that it's known name, or print error
            pass

    return dbnd_environ


def parse_and_build_config_store(
    source,
    config_values,
    override=False,
    auto_section_parse=False,
    set_if_not_exists_only=False,
):
    # type:(str, Mapping[str, Mapping[str, Any]], bool, bool , bool)->_ConfigStore
    """
    Read user defined values. Following format are supported:
        1. SomeTask.some_param [ParameterDefinition] : value
        2. { "section" : { "key" : "value" }}
        3 ? "SomeTask.some_param" [str]  : value
    """
    if isinstance(config_values, _ConfigStore):
        return config_values

    new_config = _ConfigStore()
    new_config.source = source
    for section, section_values in six.iteritems(config_values):
        if isinstance(section, six.string_types):
            if auto_section_parse:
                m = _SECTION_NAME_RE.match(section)
                if m:  # section contains key!
                    section, key = m.group(1), m.group(2)
                    section_values = {key: section_values}

            if not isinstance(section_values, Mapping):
                raise DatabandConfigError(
                    "can't convert '%s' to configuration " % config_values
                )
        elif isinstance(section, ParameterDefinition):
            # this is parameter ->  Spark.jars = ["jars"]
            section_values = {section.name: section_values}
            section = section.task_config_section

        else:
            raise Exception("section='%s' not supported" % section)

        new_section = new_config[section]
        for key, value in six.iteritems(section_values):
            if key in new_section:
                raise Exception(
                    "multiple definition of {section}.{key} at {config}".format(
                        section=section, key=key, config=config_values
                    )
                )
            if isinstance(key, ParameterDefinition):
                key = key.name
            if not isinstance(value, ConfigValue):
                value = ConfigValue(
                    value=value,
                    source=source,
                    require_parse=False,
                    override=override,
                    set_if_not_exists_only=set_if_not_exists_only,
                )
            else:
                # we can have override values without source
                if value.source is None:
                    value = attr.evolve(value, source=source)
            new_config.set_config_value(section, key, value)

    return new_config


def override(value):
    return ConfigValue(value=value, source=None, override=True)



import contextlib
import logging
import os
import shutil
import subprocess
import sys
import zipfile

from tempfile import mkdtemp
from typing import Optional

from dbnd._core.current import is_verbose
from targets import DirTarget


logger = logging.getLogger(__name__)


def _get_package_name_and_version_from_whl(whl_dir):
    all_files = os.listdir(whl_dir)
    if not all_files:
        raise Exception("No Whl files where created")

    if len(all_files) > 1:
        raise Exception("Directory has more than one whl file")

    whl_file = all_files[0]
    split_name = whl_file.split("-")
    return split_name[0], split_name[1]


def _run_command(generate_command):
    if subprocess.call(generate_command, shell=True) != 0:
        raise Exception("Failed running {} command".format(generate_command))


@contextlib.contextmanager
def _create_temp_working_dir(tmp_build_dir=None):
    clean_build_dir = False
    try:
        if not tmp_build_dir:
            tmp_build_dir = mkdtemp(prefix="dbnd-build-")
            clean_build_dir = True
        yield tmp_build_dir
    finally:
        if clean_build_dir:
            if is_verbose():  # do not clean build dir in verbose mode
                logger.info("Keeping build dir because verbose mode is on")
            else:
                logger.info("Deleting tmp directory: %s", tmp_build_dir)
                shutil.rmtree(tmp_build_dir, ignore_errors=True)


def build_package_zip(fat_py_output_dir, package_dir, tmp_build_dir=None):
    # type: (DirTarget, str, Optional[str]) -> str
    logger.info("Started building package bdist_zip file")

    with _create_temp_working_dir(tmp_build_dir) as tmp_build_dir:
        generate_project_whl(package_dir, tmp_build_dir)
        package_name, package_version = _get_package_name_and_version_from_whl(
            tmp_build_dir
        )
        fat_py_output_dir.mkdir()
        package_zip_file = os.path.join(
            str(fat_py_output_dir),
            "{}-{}-package.zip".format(package_name, package_version),
        )
        zip_dir(package_zip_file, tmp_build_dir)

        return package_zip_file


def build_third_party_zip(fat_py_output_dir, requirements_file, tmp_build_dir=None):
    # type: (DirTarget, str, Optional[str]) -> str
    logger.info("Started building third-party bdist_zip file")

    with _create_temp_working_dir(tmp_build_dir) as tmp_build_dir:
        generate_third_party_deps(requirements_file, tmp_build_dir)
        fat_py_output_dir.mkdir()
        third_party_zip_file = os.path.join(
            str(fat_py_output_dir), "third-party-deps.zip"
        )
        zip_dir(third_party_zip_file, tmp_build_dir)

        return third_party_zip_file


def build_fat_requirements_py_zip_file(
    fat_py_output_dir, package_dir, requirements_file, tmp_build_dir=None,
):
    # type: (DirTarget, str, Optional[str], str) -> str
    # similar to fat jar in java, when you can have all thirdparty deps in one file
    logger.info("Started building fat bdist_zip file")

    with _create_temp_working_dir(tmp_build_dir) as tmp_build_dir:
        generate_project_whl(package_dir, tmp_build_dir)
        package_name, package_version = _get_package_name_and_version_from_whl(
            tmp_build_dir
        )
        if requirements_file is not None:
            generate_third_party_deps(requirements_file, tmp_build_dir)
        fat_py_output_dir.mkdir()
        package_with_deps_zip = os.path.join(
            str(fat_py_output_dir),
            "{}-{}-package-with-deps.zip".format(package_name, package_version),
        )
        zip_dir(package_with_deps_zip, tmp_build_dir)
        logger.info(
            "successfully build bdist_zip fat zip file %s", package_with_deps_zip
        )

        return package_with_deps_zip


def generate_project_whl(package_dir, output_dir):
    # generating deps wheels

    # Very important to change to the working dir, otherwise, wheel creation won't work as expected
    os.chdir(package_dir)

    setup_py_path = os.path.join(package_dir, "setup.py")
    if not os.path.exists(setup_py_path):
        raise Exception("Can't find setup.py inside package dir {}".format(package_dir))

    generate_command = "{} {} bdist_wheel --dist-dir {}".format(
        sys.executable, setup_py_path, output_dir
    )

    _run_command(generate_command)


def generate_third_party_deps(requirements_file, output_dir):
    # generating deps wheels
    generate_command = "{} -m pip wheel -r {} -w {}".format(
        sys.executable, requirements_file, output_dir
    )

    _run_command(generate_command)


def zip_dir(zip_file_path, source_dir_path):
    result_zip = zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED)

    for dir_name, _, files in os.walk(source_dir_path):
        if dir_name == source_dir_path:
            for file_name in files:
                full_file_name = os.path.join(dir_name, file_name)
                temp_zip = zipfile.ZipFile(file=full_file_name, mode="r")
                for file_info in temp_zip.filelist:
                    file_content = temp_zip.read(file_info.filename)
                    result_zip.writestr(file_info, file_content)
    result_zip.close()



import logging
import time

from collections import defaultdict

from dbnd._core.task_ctrl.task_relations import TaskRelations
from dbnd._core.utils.traversing import flatten


logger = logging.getLogger(__name__)


def tensorflow_graph(task):
    from tensorboardX.src.attr_value_pb2 import AttrValue
    from tensorboardX.src.graph_pb2 import GraphDef
    from tensorboardX.src.node_def_pb2 import NodeDef
    from tensorboardX.src.step_stats_pb2 import (
        RunMetadata,
        StepStats,
        DeviceStepStats,
        NodeExecStats,
        AllocatorMemoryUsed,
    )
    from tensorboardX.src.tensor_shape_pb2 import TensorShapeProto
    from tensorboardX.src.versions_pb2 import VersionDef

    # assert isinstance(task, Task)

    tasks = task.ctrl.task_dag.subdag_tasks()
    children = defaultdict(list)
    parents = defaultdict(list)
    for task in tasks:
        children[task.task_id] = task.task_meta.get_children()
        for t_child in children[task.task_id]:
            parents[t_child.task_id].append(task)
    friendly_name = {t.task_id: t.task_af_id for t in tasks}

    def _get_name(t):
        cur_id = t.task_id
        cur_scope = [friendly_name[cur_id]]
        while parents[cur_id]:
            # right now just do simple scope, don'task take care of shared tasks
            cur_id = parents[cur_id][0].task_id
            cur_scope.append(friendly_name[cur_id])
        return "/".join(reversed(cur_scope))

    nodes = []

    # calculate scope for all tasks
    names = {}
    for task in tasks:
        names[task.task_id] = _get_name(task)

    # attrs = attrs.replace("'", ' ')  # singlequote will be escaped by tensorboard
    def _name(t, *path):
        _path = [names[t.task_id]]
        if path:
            _path.extend(list(path))
        return "/".join(_path)

    def _op_name(task):
        return _name(task, task.task_id)

    def _target_name(target):
        return _name(task.source_task, target.name or "output")

    for task in tasks:
        tr = task.ctrl.relations
        assert isinstance(tr, TaskRelations)

        task_inputs = []
        op_name = _op_name(task)

        for target in flatten(tr.task_inputs_user):
            if not target.owner:
                continue
            task_inputs.append(_target_name(target.source_task, target))

        for target in flatten(tr.task_outputs_user):
            if not target.owner:
                continue

            output_inputs = [op_name]  # current task
            if target.source_task != task:
                real_output = _target_name(target.source_task, target)
                output_inputs.append(real_output)
                task_inputs.append(real_output)
            node_info = {
                "name": _target_name(task, target),
                "op": "output",
                "task_inputs": [op_name],
                "attrs": {"path": AttrValue(s=str(target).encode(encoding="utf_8"))},
            }
            # node_info['outputsize'] = []
            nodes.append(node_info)

            attrs = {}
        for k, v in task._params.get_params_serialized():
            attrs[k] = AttrValue(s=v.encode(encoding="utf_8"))
        node_info = {
            "name": op_name,
            "op": task.task_name,
            "task_inputs": task_inputs,
            "attrs": attrs,
        }

        # node_info['outputsize'] = []
        nodes.append(node_info)

    # mapping = {}
    # for n in nodes:
    #     mapping[n['name']] = scope[n['name']] + '/' + \
    #                          n['op'] + '_' + n['name']
    # for n in nodes:
    #     n['name'] = mapping[n['name']]
    #     for i, s in enumerate(n['task_inputs']):
    #         n['task_inputs'][i] = mapping[s]

    node_stats = []
    graph_nodes = []
    for node in nodes:
        attrs = node.get("attrs", {})
        if "outputsize" in node.keys():
            shapeproto = TensorShapeProto(
                dim=[TensorShapeProto.Dim(size=d) for d in node["outputsize"]]
            )
            attrs["_output_shapes"] = AttrValue(
                list=AttrValue.ListValue(shape=[shapeproto])
            )

        if "exec_stats" in node:
            node_stats.append(
                NodeExecStats(
                    node_name=node["name"],
                    all_start_micros=int(time.time() * 1e7),
                    all_end_rel_micros=42,
                    memory=[
                        AllocatorMemoryUsed(
                            allocator_name="cpu",
                            total_bytes=19950829,
                            peak_bytes=19950829,
                            live_bytes=19950829,
                        )
                    ],
                )
            )

        graph_nodes.append(
            NodeDef(
                name=node["name"], op=node["op"], input=node["task_inputs"], attr=attrs
            )
        )

    stepstats = RunMetadata(
        step_stats=StepStats(
            dev_stats=[DeviceStepStats(device="/device:CPU:0", node_stats=node_stats)]
        )
    )
    return GraphDef(node=graph_nodes, versions=VersionDef(producer=22)), stepstats


def save_task_graph(task, comment="", log_dir=None):
    from tensorboardX import SummaryWriter

    try:
        with SummaryWriter(log_dir=log_dir, comment=comment) as w:
            w.file_writer.add_graph(tensorflow_graph(task))
    except Exception:
        logger.exception("Failed to save tensorflow graph")



import logging

from datetime import datetime
from typing import List, Optional

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import config


logger = logging.getLogger(__name__)


class RunConfig(config.Config):
    """Databand's per run settings (e.g. execution date)"""

    _conf__task_family = "run"

    # on none generated at DatabandRun
    name = parameter.value(default=None, description="Specify run name")[str]

    description = parameter.value(default=None, description="Specify run description")[
        Optional[str]
    ]

    parallel = parameter(default=None)[bool]
    task_executor_type = parameter(
        default=None,
        description="Alternate executor type: "
        " local/airflow_inprocess/airflow_multiprocess_local/airflow_kubernetes,"
        "  see docs for more options",
    )[str]

    submit_driver = parameter(
        description="override env.submit_driver for specific environment"
    ).none[bool]
    submit_tasks = parameter(
        description="override env.submit_tasks for specific environment"
    ).none[bool]

    enable_airflow_kubernetes = parameter(
        default=True,
        description="Enable use of kubernetes executor for kubebernetes engine submission",
    )[bool]

    execution_date = parameter(default=None, description="Override execution date")[
        datetime
    ]

    # Execution specific
    id = parameter(default=None, description="The list of task ids to run")[List[str]]
    selected_tasks_regex = parameter(
        default=None, description="Run only specified tasks (regular expresion)"
    )[List[str]]

    ignore_dependencies = parameter(
        description="The regex to filter specific task_ids"
    ).value(False)
    ignore_first_depends_on_past = parameter(
        description="The regex to filter specific task_ids"
    ).value(False)

    pool = parameter(default=None, description="Resource pool to use")[str]

    donot_pickle = parameter(
        description="Do not attempt to pickle the DAG object to send over "
        "to the workers, just tell the workers to run their version "
        "of the code."
    ).value(False)

    mark_success = parameter(
        description="Mark jobs as succeeded without running them"
    ).value(False)
    skip_completed = parameter(
        description="Mark jobs as succeeded without running them"
    ).value(True)
    fail_fast = parameter(
        description="Skip all remaining tasks if a task has failed"
    ).value(True)
    enable_prod = parameter(description="Enable production tasks").value(False)
    is_archived = parameter(description="Save this run in the archive").value(False)

    heartbeat_interval_s = parameter(
        description="How often a run should send a heartbeat to the server. Set -1 to disable"
    )[int]
    heartbeat_timeout_s = parameter(
        description="How old can a run's last heartbeat be before we consider it failed. Set -1 to disable"
    )[int]
    heartbeat_sender_log_to_file = parameter(
        description="create a separate log file for the heartbeat sender and don't log the run process stdout"
    )[bool]
    open_web_tracker_in_browser = parameter(
        description="If True, open web tracker in browser during task run."
    ).value(False)

    enable_concurent_sqlite = parameter(
        description="Enable concurrent execution with sqlite db (use only for debug!)"
    ).value(False)

    interactive = parameter(
        default=False,
        description="When submitting driver to remote execution keep tracking of submitted process and wait for completion",
    )[bool]

    skip_completed_on_run = parameter(default=True).help(
        "Should dbnd task check that task is completed and mark it as resued on task execution"
    )[bool]

    validate_task_inputs = parameter(default=True).help(
        "Should dbnd task check that all input files exist"
    )[bool]

    validate_task_outputs = parameter(default=True).help(
        "Should dbnd task check that all outputs exist after task has been executed"
    )[bool]

    validate_task_outputs_on_build = parameter(default=False).help(
        "Should dbnd task check that there are no incomplete outputs before task executes"
    )[bool]

    tracking_with_cache = parameter(default=False).help(
        "Should dbnd cache results during tracking"
    )[bool]

    pipeline_band_only_check = parameter(default=False).help(
        "When checking if pipeline is completed, check only if the band file exist (skip the tasks)"
    )[bool]

    task_complete_parallelism_level = parameter(default=1).help(
        "Number of threads to use when checking if tasks are already complete"
    )[int]



from __future__ import absolute_import

import logging
import typing

import six

from dbnd._core.errors import friendly_error
from targets.marshalling.file import (
    ObjJsonMarshaller,
    ObjPickleMarshaller,
    ObjYamlMarshaller,
    StrLinesMarshaller,
    StrMarshaller,
)
from targets.marshalling.marshaller_ctrl import MarshallerCtrl
from targets.target_config import FileFormat
from targets.types import DataList
from targets.values import get_value_type_of_type
from targets.values.version_value import VersionStr


logger = logging.getLogger(__name__)

MARSHALERS = {}


def register_basic_data_marshallers():
    register_marshallers(
        object,
        {
            FileFormat.txt: StrMarshaller(),
            FileFormat.pickle: ObjPickleMarshaller(),
            FileFormat.json: ObjJsonMarshaller(),
            FileFormat.yaml: ObjYamlMarshaller(),
        },
    )
    register_marshallers(
        typing.List,
        {
            FileFormat.txt: StrLinesMarshaller(),
            FileFormat.csv: StrLinesMarshaller(),
            FileFormat.json: ObjJsonMarshaller(),
            FileFormat.yaml: ObjYamlMarshaller(),
            FileFormat.pickle: ObjPickleMarshaller(),
        },
    )
    register_marshallers(
        str,
        {
            FileFormat.txt: StrMarshaller(),
            FileFormat.csv: StrMarshaller(),
            FileFormat.pickle: ObjPickleMarshaller(),
            FileFormat.json: ObjJsonMarshaller(),
        },
    )

    MARSHALERS[VersionStr] = MARSHALERS[str]

    list_marshalers = MARSHALERS[typing.List]
    for t in [typing.List[object], typing.List[str], DataList, DataList[str]]:
        MARSHALERS[t] = list_marshalers

    try:
        import numpy
        from targets.marshalling.numpy import (
            NumpyArrayMarshaller,
            NumpyArrayPickleMarshaler,
        )
    except ImportError:
        pass

    try:
        import numpy as np
        import pandas as pd

        from targets.marshalling.pandas import (
            DataFrameDictToHdf5,
            DataFrameToCsv,
            DataFrameToExcel,
            DataFrameToFeather,
            DataFrameToHdf5,
            DataFrameToJson,
            DataFrameToParquet,
            DataFrameToPickle,
            DataFrameToTable,
            DataFrameToTsv,
        )

        register_marshallers(
            pd.DataFrame,
            {
                FileFormat.txt: DataFrameToCsv(),
                FileFormat.csv: DataFrameToCsv(),
                FileFormat.table: DataFrameToTable(),
                FileFormat.parquet: DataFrameToParquet(),
                FileFormat.feather: DataFrameToFeather(),
                FileFormat.hdf5: DataFrameToHdf5(),
                FileFormat.pickle: DataFrameToPickle(),
                FileFormat.json: DataFrameToJson(),
                FileFormat.tsv: DataFrameToTsv(),
                FileFormat.excel: DataFrameToExcel(),
            },
        )
        register_marshallers(
            pd.Series,
            {
                FileFormat.csv: DataFrameToCsv(series=True),
                FileFormat.table: DataFrameToTable(),
                FileFormat.parquet: DataFrameToParquet(),
                FileFormat.feather: DataFrameToFeather(),
                FileFormat.hdf5: DataFrameToHdf5(),
                FileFormat.pickle: DataFrameToPickle(),
                FileFormat.json: DataFrameToJson(),
            },
        )
        register_marshallers(
            np.ndarray,
            {
                FileFormat.numpy: NumpyArrayMarshaller(),
                FileFormat.pickle: NumpyArrayPickleMarshaler(),
            },
        )
        register_marshallers(
            typing.Dict[str, pd.DataFrame], {FileFormat.hdf5: DataFrameDictToHdf5()}
        )
    except ImportError:
        pass
    try:
        from matplotlib import figure
        from targets.marshalling.matplotlib import MatplotlibFigureMarshaller

        register_marshallers(
            figure.Figure,
            {
                FileFormat.png: MatplotlibFigureMarshaller(),
                FileFormat.pdf: MatplotlibFigureMarshaller(),
                FileFormat.pickle: ObjPickleMarshaller(),
            },
        )
    except ImportError:
        pass


def _marshaller_options_message(value_type, value_options, object_options):
    value_options = set(value_options.keys())
    object_options = set(object_options.keys())
    object_options.difference_update(value_options)
    msg = "For '{type}' you can use {value_options}.".format(
        value_options=",".join(sorted(value_options)), type=value_type.type
    )
    if object_options:
        msg += (
            "Default marshallers (Type[object]) "
            "will be used for formats {object_options}".format(
                object_options=",".join(sorted(object_options))
            )
        )
    return msg


def get_marshaller_ctrl(target, value_type, config=None):
    config = config or target.config

    value_type = get_value_type_of_type(value_type, inline_value_type=True)
    marshaller_options = MARSHALERS.get(value_type.type, {})
    object_marshaller_options = MARSHALERS.get(object)

    target_file_format = config.format
    if target_file_format is None:
        target_file_format = FileFormat.txt
        logger.warning("Can not find file type of '%s', assuming it's .txt", target)
        # raise friendly_error.targets.no_format(
        #     target=target,
        #     options_message=_marshaller_options_message(
        #         value_type=value_type,
        #         value_options=marshaller_options,
        #         object_options=object_marshaller_options,
        #     ),
        # )

    marshaller = marshaller_options.get(target_file_format)
    if not marshaller:
        # let's try to check marshallers for object
        marshaller = object_marshaller_options.get(target_file_format)
        if not marshaller:
            raise friendly_error.targets.no_marshaller(
                target=target,
                value_type=value_type,
                config=config,
                options_message=_marshaller_options_message(
                    value_type=value_type,
                    value_options=marshaller_options,
                    object_options=object_marshaller_options,
                ),
            )
        # our value type is object!
        # may be we should print warning here?
        value_type = get_value_type_of_type(object)
    return MarshallerCtrl(target=target, marshaller=marshaller, value_type=value_type)


def register_marshaller(value_type, file_format, marshaller_cls):
    MARSHALERS.setdefault(value_type, {})[file_format] = marshaller_cls


def register_marshallers(value_type, marshaller_dict):
    marshaller_value_type = MARSHALERS.setdefault(value_type, {})
    for file_format, marshaller_cls in six.iteritems(marshaller_dict):
        marshaller_value_type[file_format] = marshaller_cls



import logging
import typing

from typing import List

from dbnd._core.run.run_ctrl import RunCtrl


if typing.TYPE_CHECKING:
    from dbnd._core.tracking.backends import TrackingStore
    from dbnd._core.task_run.task_run import TaskRun

logger = logging.getLogger(__name__)


class RunTracker(RunCtrl):
    def __init__(self, run, tracking_store):
        super(RunTracker, self).__init__(run)
        self.databand_url = self.settings.core.databand_url
        self.tracking_store = tracking_store  # type: TrackingStore

        if self.databand_url:
            self.run_url = "{databand_url}/app/jobs/{job_name}/{run_uid}".format(
                databand_url=self.databand_url,
                job_name=self.run.job_name,
                run_uid=self.run.run_uid,
            )
        else:
            self.run_url = None

    # Following handlers only works for Databand RUN, not for the specific task!
    def init_run(self):
        """
        runs for the whole dag
        we call it only for the root dag
        """
        if not self.run.is_tracked:
            return
        self.tracking_store.init_run(run=self.run)
        logger.info("Run tracking info has been committed.")

    def set_run_state(self, state):
        if not self.run.is_tracked:
            return
        self.tracking_store.set_run_state(run=self.run, state=state)

    def add_task_runs(self, task_runs):
        if not self.run.is_tracked:
            return
        self.tracking_store.add_task_runs(run=self.run, task_runs=task_runs)

    def set_task_run_states(self, task_runs):
        # type: (List[TaskRun]) -> None
        if not self.run.is_tracked:
            return
        self.tracking_store.set_task_run_states(task_runs=task_runs)



from tzlocal import get_localzone

from dbnd._core.utils.timezone import make_aware
from dbnd._vendor import click


class TZAwareDateTime(click.DateTime):
    """The TZAwareDateTime type converts date strings into `datetime` objects.

    The format strings which are checked are configurable, but default to some
    common (TIMEZONE AWARE) ISO 8601 formats.

    When specifying *DateTime* formats, you should only pass a list or a tuple.
    Other iterables, like generators, may lead to surprising results.

    The format strings are processed using ``datetime.strptime``, and this
    consequently defines the format strings which are allowed.

    Parsing is tried using each format, in order, and the first format which
    parses successfully is used.

    :param formats: A list or tuple of date format strings, in the order in
                    which they should be tried. Defaults to
                    ``'%Y-%m-%d'``, ``'%Y-%m-%dT%H:%M:%S'``,
                    ``'%Y-%m-%d %H:%M:%S'``.
    """
    name = 'tzawaredatetime'

    def _try_to_convert_date(self, value, format):
        dt = super(TZAwareDateTime, self)._try_to_convert_date(value, format)
        if dt:
            return make_aware(dt, get_localzone())

    def __repr__(self):
        return 'TZAwareDateTime'



def dont_track(obj):
    """
    will not track obj. support airflow operators, DAGs, and functions.
    allows excluding from tracking when using dbnd-airflow-auto-tracking, track_modules, and track_dag.
    can be used as a function and decorator as shown in examples below.

    Usage examples:
        dag = DAG()
        with dag:
            operator = PythonOperator()

        dont_track(operator)
        dont_track(dag)

        @dont_track
        def f():
            pass

    """
    obj._dont_track = True
    return obj


def should_not_track(obj):
    if not hasattr(obj, "_dont_track"):
        return False
    return obj._dont_track



# we need to import databand module before airflow, otherwise we will not get airflow_bome

import cProfile
import logging
import os
import random
import time

from datetime import datetime

import pytest

from _pytest.fixtures import fixture

from dbnd import config
from dbnd._core.configuration.environ_config import set_dbnd_unit_test_mode
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.context.databand_context import DatabandContext, new_dbnd_context
from dbnd._core.log.config import configure_basic_logging
from dbnd._core.task_build.task_registry import tmp_dbnd_registry
from dbnd._core.task_build.task_signature import TASK_ID_INVALID_CHAR_REGEX


logger = logging.getLogger(__name__)


def pytest_configure(config):
    configure_basic_logging(None)
    set_dbnd_unit_test_mode()
    dbnd_bootstrap()


def pytest_addoption(parser):
    # group = parser.getgroup("dnbd")
    # group._addoption(
    #     "--reuse-db",
    #     action="store_true",
    #     dest="reuse_db",
    #     default=False,
    #     help="Re-use the testing DB if it already exists, "
    #     "and do not remove it when the test finishes.",
    # )
    pass


def pytest_load_initial_conftests(early_config, parser, args):
    early_config.addinivalue_line("markers", "dbnd(: Mark the test as using dbnd")

    early_config._dbnd_report_header = None
    early_config._dbnd_report_header = "Databand: %s" % ("")


def pytest_report_header(config):
    if hasattr(config, "_dbnd_report_header") and config._dbnd_report_header:
        return [config._dbnd_report_header]


def _run_name_for_test_request(request):
    return TASK_ID_INVALID_CHAR_REGEX.sub("_", request.node.name)


@pytest.fixture(scope="session")
def databand_test_db():
    # init_test_db()
    # we are going to initialize db at databand context
    pass


@pytest.fixture
def databand_context_kwargs():
    return {}


@pytest.fixture(scope="session")
def databand_config():
    config.load_system_configs()
    return config


@pytest.fixture
def databand_test_context(
    request, tmpdir, databand_context_kwargs, databand_config
):  # type: (...) -> DatabandContext

    test_config = {
        "run": {
            "name": _run_name_for_test_request(request),
            "heartbeat_interval_s": -1,
        },
        "local": {"root": str(tmpdir.join("local_root"))},
    }
    with config(test_config, source="databand_test_context"), new_dbnd_context(
        **databand_context_kwargs
    ) as t:
        yield t


@pytest.fixture
def databand_clean_register_for_every_test():
    with tmp_dbnd_registry() as r:
        yield r


@pytest.fixture(autouse=True)
def databand_pytest_env(
    databand_test_db, databand_test_context, databand_clean_register_for_every_test
):
    """
    all required fixtures to run datband related tests
    add it to your [pytest]usefixtures
    """
    return True


@pytest.fixture()
def mock_target_fs():
    from targets.fs.mock import MockFileSystem

    MockFileSystem.instance.clear()
    yield MockFileSystem.instance
    MockFileSystem.instance.clear()


@fixture
def cProfile_benchmark():
    perf_stats_dir = "/tmp/benchmark/"

    def _benchmark(fn, *args, **kwargs):
        unique_id = datetime.now().strftime(
            "%Y%m%d-%H%M%S-" + str(random.randint(0, 10000))
        )
        # with perf_trace(os.path.join(perf_stats_dir, "p1_%s.stats" % (unique_id))):
        file = os.path.join(perf_stats_dir, "benchmark_%s.stats" % (unique_id))
        start = time.time()
        pr = cProfile.Profile()
        pr.enable()

        result = fn(*args, **kwargs)
        done = time.time()
        pr.disable()

        file = os.path.abspath(file)
        if not os.path.exists(os.path.dirname(file)):
            os.makedirs(os.path.dirname(file))

        pr.dump_stats(file)
        logging.warning("Performance report saved at %s", file)
        elapsed = done - start
        logger.warning("Time Elapsed for %s: %s", fn, elapsed)
        return result

    return _benchmark



import logging
import typing

from uuid import UUID


if typing.TYPE_CHECKING:
    from typing import List, Optional, Union

    from targets.base_target import Target
    from targets.value_meta import ValueMeta

    from dbnd.api.tracking_api import (
        InitRunArgs,
        AirflowTaskInfo,
        LogTargetArgs,
    )
    from dbnd._core.constants import (
        DbndTargetOperationType,
        DbndTargetOperationStatus,
        UpdateSource,
    )
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd._core.run.databand_run import DatabandRun

logger = logging.getLogger(__name__)


class TrackingStore(object):
    def init_scheduled_job(self, scheduled_job, update_existing):
        pass

    def init_run(self, run):
        # type: (DatabandRun) -> List[int]
        pass

    def init_run_from_args(self, init_args):
        # type: (InitRunArgs) -> List[int]
        pass

    def set_run_state(self, run, state, timestamp=None):
        pass

    def set_task_reused(self, task_run):
        pass

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        pass

    def set_task_run_states(self, task_runs):
        pass

    def set_unfinished_tasks_state(self, run_uid, state):
        pass

    def save_task_run_log(self, task_run, log_body, local_log_path=None):
        pass

    def save_external_links(self, task_run, external_links_dict):
        pass

    def log_histograms(self, task_run, key, value_meta, timestamp):
        pass

    def log_metrics(self, task_run, metrics):
        pass

    def log_artifact(self, task_run, name, artifact, artifact_target):
        pass

    def add_task_runs(self, run, task_runs):
        pass

    def heartbeat(self, run_uid):
        pass

    def update_task_run_attempts(self, task_run_attempt_updates):
        pass

    def save_airflow_task_infos(self, airflow_task_infos, source, base_url):
        # type: (List[AirflowTaskInfo], UpdateSource, str) -> None
        pass

    def log_target(
        self,
        task_run,  # type: TaskRun
        target,  # type: Union[Target, str]
        target_meta,  # type: ValueMeta
        operation_type,  # type: DbndTargetOperationType
        operation_status,  # type: DbndTargetOperationStatus
        param_name=None,  # type: Optional[str]
        task_def_uid=None,  # type: Optional[UUID]
    ):  # type: (...) -> None
        pass

    def log_targets(self, targets_info):
        # type: (List[LogTargetArgs]) -> None
        pass

    def is_ready(self):
        # type: () -> bool
        pass

    def save_airflow_monitor_data(self, **kwargs):
        pass



#  << should be run before import to airflow >>
# otherwise airflow.configuration will fail
# fix AIRFLOW_HOME for all runs

import os

from configparser import ConfigParser
from contextlib import contextmanager

from dbnd._core.configuration.project_env import (
    _init_windows_python_path,
    _is_init_mode,
)
from dbnd._core.utils.basics.environ_utils import (
    env,
    environ_enabled,
    environ_int,
    set_env_dir,
    set_on,
)
from dbnd._core.utils.basics.path_utils import abs_join, relative_path


_MARKER_FILES = ["databand.cfg", "project.cfg", "databand-system.cfg"]
PARAM_ENV_TEMPLATE = "DBND__{S}__{K}"

ENV_DBND__DISABLED = "DBND__DISABLED"
ENV_DBND__TRACKING = (
    "DBND__TRACKING"  # implicit DBND tracking ( on any @task/log_ call)
)
ENV_DBND__VERBOSE = "DBND__VERBOSE"  # VERBOSE
ENV_DBND__UNITTEST_MODE = "DBND__UNITTEST"
ENV_DBND_QUIET = "DBND__QUIET"

ENV_DBND_HOME = "DBND_HOME"
ENV_DBND_SYSTEM = "DBND_SYSTEM"
ENV_DBND_LIB = "DBND_LIB"
ENV_DBND_CONFIG = "DBND_CONFIG"  # extra config for DBND

ENV_DBND__DEBUG_INIT = "DBND__DEBUG_INIT"

ENV_DBND__USER_PRE_INIT = "DBND__USER_PRE_INIT"  # run on user init
ENV_DBND__NO_MODULES = "DBND__NO_MODULES"  # do not auto-load user modules

ENV_DBND__NO_TABLES = "DBND__NO_TABLES"  # do not print fancy tables
ENV_DBND__SHOW_STACK_ON_SIGQUIT = "DBND__SHOW_STACK_ON_SIGQUIT"
ENV_DBND__OVERRIDE_AIRFLOW_LOG_SYSTEM_FOR_TRACKING = (
    "DBND__OVERRIDE_AIRFLOW_LOG_SYSTEM_FOR_TRACKING"
)
ENV_DBND__DISABLE_AIRFLOW_SUBDAG_TRACKING = "DBND__DISABLE_AIRFLOW_SUBDAG_TRACKING"

ENV_DBND_USER = "DBND_USER"
ENV_DBND_ENV = "DBND_ENV"

# DBND RUN info variables
SCHEDULED_DAG_RUN_ID_ENV = "SCHEDULED_DAG_RUN_ID"
SCHEDULED_DATE_ENV = "SCHEDULED_DATE"
SCHEDULED_JOB_UID_ENV = "SCHEDULED_JOB_UID"
DBND_ROOT_RUN_UID = "DBND_ROOT_RUN_UID"
DBND_ROOT_RUN_TRACKER_URL = "DBND_ROOT_RUN_TRACKER_URL"
DBND_PARENT_TASK_RUN_UID = "DBND_PARENT_TASK_RUN_UID"
DBND_PARENT_TASK_RUN_ATTEMPT_UID = "DBND_PARENT_TASK_RUN_ATTEMPT_UID"
DBND_RUN_SUBMIT_UID = "DBND_SUBMIT_UID"
DBND_RUN_UID = "DBND_RUN_UID"
DBND_RESUBMIT_RUN = "DBND_RESUBMIT_RUN"
DBND_TASK_RUN_ATTEMPT_UID = "DBND_TASK_RUN_ATTEMPT_UID"
DBND_MAX_CALLS_PER_RUN = "DBND_MAX_CALL_PER_FUNC"
ENV_DBND_DISABLE_SCHEDULED_DAGS_LOAD = "DBND_DISABLE_SCHEDULED_DAGS_LOAD"

ENV_DBND__ENV_MACHINE = "DBND__ENV_MACHINE"
ENV_DBND__ENV_IMAGE = "DBND__ENV_IMAGE"
ENV_DBND__CORE__PLUGINS = "DBND__CORE__PLUGINS"

ENV_SHELL_COMPLETION = "_DBND_COMPLETE"

ENV_DBND_FIX_PYSPARK_IMPORTS = "DBND__FIX_PYSPARK_IMPORTS"
ENV_DBND__DISABLE_PLUGGY_ENTRYPOINT_LOADING = "DBND__DISABLE_PLUGGY_ENTRYPOINT_LOADING"

ENV_DBND__AUTO_TRACKING = "DBND__AUTO_TRACKING"

DEFAULT_MAX_CALLS_PER_RUN = 100

ENV_DBND_TRACKING_ATTEMPT_UID = "DBND__TRACKING_ATTEMPT_UID"

_DBND_DEBUG_INIT = environ_enabled(ENV_DBND__DEBUG_INIT)
_databand_package = relative_path(__file__, "..", "..")


def is_databand_enabled():
    return not get_dbnd_project_config().disabled


def disable_databand():
    get_dbnd_project_config().disabled = True


def set_dbnd_unit_test_mode():
    set_on(ENV_DBND__UNITTEST_MODE)  # bypass to subprocess
    get_dbnd_project_config().unit_test_mode = True


def get_max_calls_per_func():
    return get_dbnd_project_config().max_calls_per_run


# User setup configs
def get_dbnd_environ_config_file():
    return os.environ.get(ENV_DBND_CONFIG, None)


def get_user_preinit():
    return os.environ.get(ENV_DBND__USER_PRE_INIT, None)


def in_quiet_mode():
    """
    quiet mode was made for the scheduler to silence the launcher runners.
    Don't want this flag to propagate into the actual scheduled cmd
    """
    return get_dbnd_project_config().quiet_mode


def is_unit_test_mode():
    return get_dbnd_project_config().unit_test_mode


def spark_tracking_enabled():
    return environ_enabled("DBND__ENABLE__SPARK_CONTEXT_ENV")


def should_fix_pyspark_imports():
    return environ_enabled(ENV_DBND_FIX_PYSPARK_IMPORTS)


_project_config = None  # type: Optional[DbndProjectConfig]


def get_dbnd_project_config():
    global _project_config
    if not _project_config:
        # initialize dbnd home first
        _project_config = DbndProjectConfig()
        _initialize_dbnd_home()
    return _project_config


def get_dbnd_custom_config():
    try:
        import dbnd_custom_config

        return dbnd_custom_config.get_config_file_path()
    except Exception:
        return ""


def reset_dbnd_project_config():
    global _project_config
    _project_config = None


@contextmanager
def new_dbnd_project_config_context(**environment):
    """
    Create new environment context and reset the project int it.
    """
    with env(**environment):
        reset_dbnd_project_config()
        yield get_dbnd_project_config()

    reset_dbnd_project_config()


class DbndProjectConfig(object):
    """
    very basic environment config!
    """

    def __init__(self):
        # IF FALSE  - we will not modify decorated @task code
        self._disabled = environ_enabled(ENV_DBND__DISABLED, False)
        self.unit_test_mode = environ_enabled(ENV_DBND__UNITTEST_MODE)

        self.max_calls_per_run = environ_int(
            DBND_MAX_CALLS_PER_RUN, DEFAULT_MAX_CALLS_PER_RUN
        )

        self.shell_cmd_complete_mode = ENV_SHELL_COMPLETION in os.environ
        self.quiet_mode = (
            os.environ.pop(ENV_DBND_QUIET, None) is not None
            or self.shell_cmd_complete_mode
        )

        self.is_no_modules = environ_enabled(ENV_DBND__NO_MODULES)
        self.disable_pluggy_entrypoint_loading = environ_enabled(
            ENV_DBND__DISABLE_PLUGGY_ENTRYPOINT_LOADING
        )
        self.is_sigquit_handler_on = environ_enabled(ENV_DBND__SHOW_STACK_ON_SIGQUIT)

        self._verbose = environ_enabled(ENV_DBND__VERBOSE)

        self._dbnd_tracking = environ_enabled(ENV_DBND__TRACKING, default=None)

        self._airflow_context = False
        self._inline_tracking = None

        self.disable_inline = False
        self.airflow_auto_tracking = environ_enabled(
            ENV_DBND__AUTO_TRACKING, default=True
        )

    @property
    def disabled(self):
        return self._disabled

    @disabled.setter
    def disabled(self, value):
        set_on(ENV_DBND__DISABLED)
        self._disabled = value

    def airflow_context(self):
        if not self._airflow_context:
            from dbnd._core.inplace_run.airflow_dag_inplace_tracking import (
                try_get_airflow_context,
            )

            self._airflow_context = try_get_airflow_context()
        return self._airflow_context

    def is_tracking_mode(self):
        if self.disabled:
            return False

        if self._dbnd_tracking is None:
            return bool(self.airflow_context())

        return self._dbnd_tracking

    def is_verbose(self):
        from dbnd._core.current import try_get_databand_context

        context = try_get_databand_context()
        if context and getattr(context, "system_settings", None):
            if context.system_settings.verbose:
                return True

        return self._verbose

    def dbnd_home(self):
        return os.environ.get(ENV_DBND_HOME) or os.curdir

    def dbnd_lib_path(self, *path):
        return abs_join(_databand_package, *path)

    def dbnd_config_path(self, *path):
        return self.dbnd_lib_path("conf", *path)

    def dbnd_system_path(self, *path):
        dbnd_system = os.environ.get(ENV_DBND_SYSTEM) or self.dbnd_home()
        return abs_join(dbnd_system, *path)

    def dbnd_project_path(self, *path):
        return abs_join(self.dbnd_home(), *path)

    def validate_init(self):
        _debug_init_print("Successfully created dbnd project config")


def _debug_init_print(msg):
    if _DBND_DEBUG_INIT:
        print("DBND INIT: %s" % msg)


class DatabandHomeError(Exception):
    pass


def _find_project_by_import():
    """
    check if we can have project marker file by import it
    """
    try:
        import _databand_project

        return abs_join(_databand_project.__file__, "..")
    except ImportError:
        _debug_init_print("Can't import `_databand_project` marker.")
    return None


def _process_cfg(folder):
    # dbnd home is being pointed inside [databand] in 'config' files
    found_dbnd_home = False
    config_file = None

    config_files = ["tox.ini", "setup.cfg"]
    for file in config_files:
        config_path = os.path.join(folder, file)
        try:
            parser = ConfigParser()
            parser.read(config_path)

            config_root, config_name = os.path.split(config_path)
            source = os.path.basename(config_path)
            if not parser.has_section("databand"):
                continue

            for config_key in ["dbnd_home", "dbnd_system", "dbnd_config"]:
                # TODO: hidden magic, do we need these setters?
                if not parser.has_option("databand", config_key):
                    continue
                config_value = parser.get("databand", config_key)
                config_value = os.path.abspath(os.path.join(config_root, config_value))
                set_env_dir(config_key, config_value)
                _debug_init_print("%s: %s=%s" % (source, config_key, config_value))

        except Exception as ex:
            print("Failed to process %s: %s" % (config_path, ex))

    return found_dbnd_home, config_file


def _has_marker_file(folder):
    # dbnd home is where 'marker' files are located in
    for file in _MARKER_FILES:
        file_path = os.path.join(folder, file)
        if os.path.exists(file_path):
            return folder, file_path

    return False, None


def __find_dbnd_home_at(folder):
    dbnd_home, config_file = _process_cfg(folder)
    if dbnd_home:
        _debug_init_print(
            "Found dbnd home by at %s, by config file: %s" % (dbnd_home, config_file)
        )
        return dbnd_home

    dbnd_home, marker_file = _has_marker_file(folder)
    if dbnd_home:
        _debug_init_print(
            "Found dbnd home by at %s, by marker file: %s" % (dbnd_home, marker_file)
        )
        return dbnd_home

    _debug_init_print("dbnd home was not found at %s" % folder)
    return False


def _find_and_set_dbnd_home():
    # falling back to simple version
    if _is_init_mode():
        dbnd_home = os.path.abspath(".")
        print("Initializing new dbnd environment, using %s as DBND_HOME" % dbnd_home)
        set_env_dir(ENV_DBND_HOME, dbnd_home)
        return True

    # looking for dbnd project folder to be set as home
    project_folder = _find_project_by_import()
    if project_folder:
        _debug_init_print(
            "Found project folder by import from marker %s" % project_folder
        )
        # we know about project folder, let try to find "custom" configs in it
        dbnd_home = __find_dbnd_home_at(project_folder)
        if not dbnd_home:
            _debug_init_print(
                "No markers at %s! setting dbnd_home to %s" % (dbnd_home, dbnd_home)
            )
            dbnd_home = project_folder

        set_env_dir(ENV_DBND_HOME, dbnd_home)
        return True

    # traversing all the way up to until finding relevant anchor files
    cur_dir = os.path.normpath(os.getcwd())
    cur_dir_split = cur_dir.split(os.sep)
    cur_dir_split_reversed = reversed(list(enumerate(cur_dir_split)))
    _debug_init_print(
        "Trying to find dbnd_home by traversing up to the root folder starting at %s"
        % cur_dir
    )

    for idx, cur_folder in cur_dir_split_reversed:
        cur_path = os.path.join("/", *cur_dir_split[1 : (idx + 1)])

        dbnd_system_file = os.path.join(cur_path, ".dbnd", "databand-system.cfg")
        if os.path.exists(dbnd_system_file):
            set_env_dir(ENV_DBND_HOME, cur_path)
            return True

        dbnd_home = __find_dbnd_home_at(cur_path)
        if dbnd_home:
            set_env_dir(ENV_DBND_HOME, dbnd_home)
            return True

    # last chance, we couldn't find dbnd project so we'll use user's home folder
    user_home = os.path.expanduser("~")
    if user_home:
        _debug_init_print("dbnd home was not found. Using user's home: %s" % user_home)
        set_env_dir(ENV_DBND_HOME, user_home)
        return True

    return False


def _env_banner():
    return "\tDBND_HOME={dbnd}\n\tDBND_SYSTEM={system}".format(
        dbnd=os.environ[ENV_DBND_HOME], system=os.environ[ENV_DBND_SYSTEM]
    )


_DBND_ENVIRONMENT = False


def _initialize_dbnd_home():
    global _DBND_ENVIRONMENT
    if _DBND_ENVIRONMENT:
        return
    _DBND_ENVIRONMENT = True

    _debug_init_print("_initialize_dbnd_home")
    project_config = get_dbnd_project_config()
    if project_config.disabled:
        _debug_init_print("databand is disabled via %s" % ENV_DBND__DISABLED)
        return

    _debug_init_print("Initializing Databand Basic Environment")
    _initialize_google_composer()
    _init_windows_python_path(_databand_package)

    # main logic
    __initialize_dbnd_home_environ()

    __initialize_airflow_home()

    if project_config.quiet_mode:
        # we should not print anything if we are in shell completion!
        import logging

        logging.getLogger().setLevel(logging.CRITICAL + 1)

    _debug_init_print(_env_banner())


def __initialize_dbnd_home_environ():
    # MAIN PART OF THE SCRIPT
    if ENV_DBND_HOME not in os.environ:
        _find_and_set_dbnd_home()

    if ENV_DBND_HOME not in os.environ:
        raise DatabandHomeError(
            "\nDBND_HOME could not be found when searching from current directory '%s' to root folder! \n "
            "Use `export DBND__DEBUG_INIT=True` to get more debug information.\n"
            "Trying fixing that issue by:\n"
            "\t 1. Explicitly set current directory to DBND HOME via: `export DBND_HOME=ROOT_OF_YOUR_PROJECT`.\n"
            "\t 2. `cd` into your project directory.\n"
            "\t 3. Create one of the following files inside current directory: [%s].\n"
            "\t 4. Run 'dbnd project-init' in current directory."
            % (os.getcwd(), ", ".join(_MARKER_FILES))
        )

    _dbnd_home = os.environ[ENV_DBND_HOME]

    if ENV_DBND_SYSTEM not in os.environ:
        dbnd_system = os.path.join(_dbnd_home, ".dbnd")

        # backward compatibility to $DBND_HOME/dbnd folder
        if not os.path.exists(dbnd_system) and os.path.exists(
            os.path.join(_dbnd_home, "dbnd")
        ):
            dbnd_system = os.path.join(_dbnd_home, "dbnd")

        os.environ[ENV_DBND_SYSTEM] = dbnd_system

    if ENV_DBND_LIB in os.environ:
        # usually will not happen
        _debug_init_print("Using DBND Library from %s" % os.environ[ENV_DBND_LIB])
    else:
        os.environ[ENV_DBND_LIB] = _databand_package


def __initialize_airflow_home():
    ENV_AIRFLOW_HOME = "AIRFLOW_HOME"

    if ENV_AIRFLOW_HOME in os.environ:
        # user settings - we do nothing
        _debug_init_print(
            "Found user defined AIRFLOW_HOME at %s" % os.environ[ENV_AIRFLOW_HOME]
        )
        return

    for dbnd_airflow_home in [
        os.path.join(os.environ[ENV_DBND_SYSTEM], "airflow"),
        os.path.join(os.environ[ENV_DBND_HOME], ".airflow"),
    ]:
        if not os.path.exists(dbnd_airflow_home):
            continue

        _debug_init_print(
            "Found airflow home folder at DBND, setting AIRFLOW_HOME to %s"
            % dbnd_airflow_home
        )
        os.environ[ENV_AIRFLOW_HOME] = dbnd_airflow_home


def _initialize_google_composer():
    if "COMPOSER_ENVIRONMENT" not in os.environ:
        return
    _debug_init_print("Initializing Google Composer Environment")

    if ENV_DBND_HOME not in os.environ:
        os.environ[ENV_DBND_HOME] = os.environ["HOME"]

    env_tracker_raise_on_error = "DBND__CORE__TRACKER_RAISE_ON_ERROR"
    if env_tracker_raise_on_error not in os.environ:
        os.environ[env_tracker_raise_on_error] = "false"


def dbnd_project_path(*path):
    return get_dbnd_project_config().dbnd_project_path(*path)



import logging

from contextlib import contextmanager
from typing import Any

from dbnd._core.errors import DatabandError, DatabandSystemError


logger = logging.getLogger(__name__)


class SingletonContext(object):
    """
    We need this object as we want to wrap __enter__ and __exit__
    function of every object with singleton management

    So inside 'with' statement we will get current object at get_instance()
    """

    _instance = None

    @classmethod
    def instance(cls):
        return cls.get_instance()

    @classmethod
    def get_instance(cls):
        """ Singleton getter """
        if not cls._instance:
            raise DatabandSystemError(
                "%s is not set, call .global_instance or .try_instance first"
                % cls.__name__
            )

        return cls._instance

    @classmethod
    def try_get_instance(cls):
        return cls._instance

    @classmethod
    def has_instance(cls):
        """ Singleton getter """
        return bool(cls._instance)

    @classmethod
    def try_instance(cls, *args, **kwargs):
        """ Singleton get or create"""
        if cls._instance is None:
            cls._instance = cls.new_context(*args, **kwargs).__enter__()
        return cls._instance

    @classmethod
    def renew_instance(cls, *args, **kwargs):
        cls._instance = None
        return cls.try_instance(*args, **kwargs)

    @classmethod
    @contextmanager
    def new_or_existing_context(cls, *args, **kwargs):
        if cls.has_instance():
            yield cls._instance
        else:
            with cls.new_context(*args, **kwargs) as new_context:
                yield new_context

    @classmethod
    @contextmanager
    def new_context(cls, *args, **kwargs):  # type: ('T', *Any, **Any)-> 'T'
        """
        Meant to be used as a context manager.
        """
        allow_override = kwargs.pop("allow_override", False)
        context = kwargs.pop("_context", None)

        orig_value = cls._instance
        if orig_value is not None:
            if not allow_override:
                raise DatabandSystemError(
                    "You are trying to create new %s out of existing context '%s', "
                    "are you sure you are allowed to do that? "
                    % (cls.__name__, cls._instance)
                )
        if context is None:
            try:
                context = cls(*args, **kwargs)
            except DatabandError:
                logger.error("Failed to create new context for %s", cls)
                raise
            except Exception:
                logger.exception("Failed to create new context for %s", cls)
                raise

        cls._instance = context

        try:
            cls._instance._on_enter()
            yield cls._instance
        finally:
            if cls._instance is not context:
                raise DatabandSystemError(
                    "Something wrong with %s context manager, "
                    "somebody has change context while we were running: "
                    "actual=%s, expected=%s" % (cls.__name__, cls._instance, context)
                )

            cls._instance._on_exit()
            cls._instance = orig_value

    @classmethod
    @contextmanager
    def context(cls, _context, **kwargs):  # type: ('T', 'T', **Any)-> 'T'
        """
        Set current context to _context
        """
        with cls.new_context(_context=_context, allow_override=True, **kwargs) as c:
            yield c

    def _on_enter(self):
        pass

    def _on_exit(self):
        pass



# -*- coding: utf-8 -*-

from dbnd import (
    Config,
    DataSourceTask,
    PipelineTask,
    PythonTask,
    Task,
    as_task,
    auto_namespace,
    band,
    config,
    config_deco,
    current,
    data,
    dbnd_context,
    dbnd_handle_errors,
    dbnd_run_start,
    dbnd_run_stop,
    hookimpl,
    namespace,
    new_dbnd_context,
    output,
    override,
    parameter,
    pipeline,
    task,
)
from dbnd.tasks import basics
from targets import _set_patches


__all__ = [
    "override",
    "config",
    "config_deco",
    "new_dbnd_context",
    "dbnd_context",
    "band",
    "pipeline",
    "task",
    "dbnd_handle_errors",
    "hookimpl",
    "dbnd_run_start",
    "dbnd_run_stop",
    "Config",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
    "Task",
    "current",
    "auto_namespace",
    "namespace",
    "as_task",
    "data",
    "output",
    "parameter",
    "basics",
    "_set_patches",
]

dbnd_config = config
# if you need advanced Tasks or Parameters please use
# databand.tasks

# shortcuts for useful objects
str(_set_patches)  # NOQA

__version__ = "0.32.6"

__title__ = "databand"
__description__ = "Machine Learning Orchestration"
__url__ = "http://www.databand.ai/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Evgeny Shulman"
__email__ = "evgeny.shulman@databand.ai"

__license__ = "Commercial Licenc e"
__copyright__ = "Copyright (c) 2018 databand.ai"



import abc
import operator

from dbnd._core.errors import friendly_error


class Validator(object):
    @property
    @abc.abstractmethod
    def description(self):
        pass

    @abc.abstractmethod
    def validate(self, parameter, value):
        pass


class NonEmptyString(Validator):
    @property
    def description(self):
        return "Non empty string"

    def validate(self, parameter, value):
        if not value:
            raise friendly_error.config.empty_string_validator(parameter)
        return value


class ChoiceValidator(Validator):
    """
    A parameter which takes two values:
        1. an instance of :class:`~collections.Iterable` and
        2. the class of the variables to convert to.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
            my_param = parameter.choices([0.1, 0.2, 0.3])[float]

    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --my-param 0.1

    Consider using :class:`~databand.EnumParameter` for a typed, structured
    alternative.  This class can perform the same role when all choices are the
    same type and transparency of parameter value on the command line is
    desired.
    """

    def __init__(self, choices):
        """
        :param function value_type: The type of the input variable, e.g. str, int,
                                  float, etc.
                                  Default: str
        :param choices: An iterable, all of whose elements are of `value_type` to
                        restrict parameter choices to.
        """

        super(ChoiceValidator, self).__init__()
        self._choices = set(choices)

    @property
    def description(self):
        return "Choices: {" + ", ".join(str(choice) for choice in self._choices) + "}"

    def validate(self, parameter, value):

        assert all(
            parameter.value_type.is_type_of(choice) for choice in self._choices
        ), "Invalid type in choices"

        if value not in self._choices:
            raise friendly_error.config.choice_error(
                parameter, value=value, choices=self._choices
            )
        return value


class NumericalValidator(Validator):
    """
    Parameter whose value is a number of the specified type, e.g. ``int`` or
    ``float`` and in the range specified.

    In the task definition, use

    .. code-block:: python

        class MyTask(dbnd.Task):
            my_param_1 = databand.NumericalParameter(
                value_type=int, min_value=-3, max_value=7) # -3 <= my_param_1 < 7
            my_param_2 = databand.NumericalParameter(
                value_type=int, min_value=-3, max_value=7, left_op=operator.lt, right_op=operator.le) # -3 < my_param_2 <= 7

    At the command line, use

    .. code-block:: console

        $ dbnd --module my_tasks MyTask --my-param-1 -3 --my-param-2 -2
    """

    def __init__(self, min_value, max_value, left_op=operator.le, right_op=operator.lt):
        """
        :param function value_type: The type of the input variable, e.g. int or float.
        :param min_value: The minimum value permissible in the accepted values
                          range.  May be inclusive or exclusive based on left_op parameter.
                          This should be the same type as value_type.
        :param max_value: The maximum value permissible in the accepted values
                          range.  May be inclusive or exclusive based on right_op parameter.
                          This should be the same type as value_type.
        :param function left_op: The comparison operator for the left-most comparison in
                                 the expression ``min_value left_op value right_op value``.
                                 This operator should generally be either
                                 ``operator.lt`` or ``operator.le``.
                                 Default: ``operator.le``.
        :param function right_op: The comparison operator for the right-most comparison in
                                  the expression ``min_value left_op value right_op value``.
                                  This operator should generally be either
                                  ``operator.lt`` or ``operator.le``.
                                  Default: ``operator.lt``.
        """
        self._min_value = min_value
        self._max_value = max_value

        self._left_op = left_op
        self._right_op = right_op
        self._permitted_range = " {left_endpoint}{min_value}, {max_value}{right_endpoint}".format(
            min_value=self._min_value,
            max_value=self._max_value,
            left_endpoint="[" if left_op == operator.le else "(",
            right_endpoint=")" if right_op == operator.lt else "]",
        )

    @property
    def description(self):
        return "permitted values: " + self._permitted_range

    def validate(self, parameter, value):
        if self._left_op(self._min_value, value) and self._right_op(
            value, self._max_value
        ):
            return value
        else:
            raise ValueError(
                "{s} is not in the set of {permitted_range}".format(
                    s=value, permitted_range=self._permitted_range
                )
            )



class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def _as_dotted_dict(**dict_obj):
    return dotdict(**dict_obj)



from dbnd._core.errors.base import (
    DatabandBuildError,
    DatabandConfigError,
    DatabandError,
    DatabandRunError,
    DatabandRuntimeError,
    DatabandSystemError,
    MissingParameterError,
    ParameterError,
    ParseParameterError,
    ParseValueError,
    TaskClassAmbigiousException,
    TaskClassNotFoundException,
    TaskValidationError,
    UnknownParameterError,
    ValueTypeError,
)
from dbnd._core.errors.errors_utils import get_help_msg, show_exc_info



import errno
import os


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid == 0:
        # According to "man 2 kill" PID 0 has a special meaning:
        # it refers to <<every process in the process group of the
        # calling process>> so we don't want to go any further.
        # If we get here it means this UNIX platform *does* have
        # a process with id 0.
        return True
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH) therefore we should never get
            # here. If we do let's be explicit in considering this
            # an error.
            raise err
    else:
        return True



import logging

from dbnd._core.current import is_verbose


logger = logging.getLogger(__name__)


def dbnd_log_debug(msg, *args, **kwargs):
    try:
        if is_verbose():
            logger.info(msg, *args, **kwargs)
        else:
            logger.debug(msg, *args, **kwargs)
    except:
        print("Failed to print dbnd info message")


def dbnd_log_info_error(msg, *args, **kwargs):
    """we show exception only in verbose mode"""
    try:
        if is_verbose():
            logger.exception(msg, *args, **kwargs)
        else:
            logger.info(msg, *args, **kwargs)
    except Exception:
        print("Failed to print dbnd error message")



from typing import List

from dbnd import task
from dbnd.tasks.py_distribution.fat_wheel_builder import (
    build_fat_wheel,
    build_wheel_zips,
)


@task
def fat_wheel_building_task(use_cached=True, output_dir=None):
    # type: (bool, str) -> str
    return build_fat_wheel(use_cached, output_dir)


@task
def wheel_zips_building_task(use_cached=True, output_dir=None):
    # type: (bool, str) -> List[str]
    return build_wheel_zips(use_cached, output_dir)



import logging
import typing

from contextlib import contextmanager
from typing import Any

import attr

from dbnd._core.constants import DbndTargetOperationStatus, DbndTargetOperationType
from dbnd._core.current import try_get_current_task_run
from dbnd._core.errors import friendly_error
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from targets import InMemoryTarget
from targets.values import get_value_type_of_obj


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.task.base_task import _BaseTask


@attr.s
class TaskAutoParamsReadWrite(object):
    task = attr.ib()  # type: _BaseTask
    auto_read = attr.ib()
    save_on_change = attr.ib()
    normalize_on_change = attr.ib()

    @contextmanager
    def auto_load_save_params(self):
        task = self.task
        original_values = task._params.get_param_values()

        if self.auto_read:
            task.load_task_runtime_values()
            original_values = task._params.get_param_values()
            task._task_auto_read_original = {p.name: v for p, v in original_values}

        dirty = self.save_on_change
        try:
            yield original_values
            # now we disable "auto read"
            task._task_auto_read_original = None
            # from here we are going to read "the value" without autoresolving

            current_values = {
                p.name: value for p, value in task._params.get_param_values()
            }
            if not self.save_on_change and not self.normalize_on_change:
                return

            changed = []
            for p, original_value in original_values:
                current_value = current_values[p.name]
                if id(original_value) != id(current_value):
                    # nothing to do original_value is the same
                    changed.append((p, original_value, current_value))

            if self.save_on_change:
                try:
                    for p, original_value, current_value in changed:
                        # TODO: implement Atomic commit
                        if p.is_output():
                            self.auto_save_param(p, original_value, current_value)
                finally:
                    for p, original_value, current_value in changed:
                        setattr(task, p.name, original_value)
                    dirty = False

            elif self.normalize_on_change:
                for p, original_value, current_value in changed:
                    try:
                        # probably we are in the band
                        # we are going just to normalize the value
                        if p.is_output():
                            from dbnd._core.utils.task_utils import to_targets

                            normalized_value = to_targets(current_value)
                        else:
                            normalized_value = p.normalize(current_value)
                        if id(normalized_value) != id(current_value):
                            setattr(task, p.name, normalized_value)
                    except Exception as ex:
                        raise friendly_error.task_build.failed_to_assign_param_value_at_band(
                            ex, p, current_value, task
                        )
        finally:
            if dirty:
                for p, original_value in original_values:
                    setattr(task, p.name, original_value)

    def auto_save_param(self, parameter, original_value, current_value):
        # type: (ParameterDefinition, Any, Any) -> None
        # it's output! we are going to save it.
        # task run doesn't always exist
        task_run = try_get_current_task_run()
        access_status = DbndTargetOperationStatus.OK
        try:
            if isinstance(original_value, InMemoryTarget):
                parameter.value_type = get_value_type_of_obj(
                    current_value, parameter.value_type
                )

            parameter.dump_to_target(original_value, current_value)
            # it's a workaround, we don't want to change parameter for outputs (dynamically)
            # however, we need proper value type to "dump" preview an other meta.
            # we will update it only for In memory targets only for now

        except Exception as ex:
            access_status = DbndTargetOperationStatus.NOK
            raise friendly_error.task_execution.failed_to_save_value_to_target(
                ex, self.task, parameter, original_value, current_value
            )
        finally:
            if task_run:
                try:
                    task_run.tracker.log_parameter_data(
                        parameter=parameter,
                        target=original_value,
                        value=current_value,
                        operation_type=DbndTargetOperationType.write,
                        operation_status=access_status,
                    )
                except Exception as ex:
                    logger.warning("Failed to log target to tracking store. %s", ex)



import logging
import typing
import uuid

from typing import Optional
from uuid import UUID

from dbnd._core.configuration.config_readers import read_from_config_files
from dbnd._core.configuration.config_store import ConfigMergeSettings
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.configuration.environ_config import is_unit_test_mode
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.errors.errors_utils import UserCodeDetector
from dbnd._core.plugin.dbnd_plugins import pm
from dbnd._core.run.databand_run import new_databand_run
from dbnd._core.settings import DatabandSystemConfig, OutputConfig, RunInfoConfig
from dbnd._core.task.task import Task
from dbnd._core.task_build.task_instance_cache import TaskInstanceCache
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd._core.utils import seven
from dbnd._core.utils.basics.load_python_module import load_python_module, run_user_func
from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.basics.singleton_context import SingletonContext
from dbnd._core.utils.timezone import utcnow
from targets.target_config import FileFormat


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun


if typing.TYPE_CHECKING:
    from typing import ContextManager

logger = logging.getLogger(__name__)


class DatabandContext(SingletonContext):
    """
    Helper for parsing command line arguments and configuration and used as part of the
    context when instantiating task objects.
    Holds state of all "config" providers, like cmd line and configuration files

    This object is managed via SingletonContext, with on_enter and on_exit functions

    When Object is built it accesses settings/config via this object
    So we can't build object till we don't get to on_enter function.
    If we don't have root  task - that means we run without command line - programmatic invocation,
    we'll create new Dag and run everything under its context

    """

    _current_context_id = 0

    # controls load of orm dags by versioned airflow
    # we will set it to true when we run airflow original commands
    def __init__(self, module=None, name="global", autoload_modules=True):
        """
        Initialize cmd line args
        """
        super(DatabandContext, self).__init__()
        # double checking on bootstrap, as we can run from all kind of locations
        # usually we should be bootstraped already as we run from cli.
        dbnd_bootstrap()

        self.name = name

        self.current_context_uid = "%s_%s" % (
            utcnow().strftime("%Y%m%d_%H%M%S"),
            str(uuid.uuid1())[:8],
        )

        self._module = module
        self.config = config

        # we are running from python notebook, let start to print to stdout
        if self.name == "interactive" or is_unit_test_mode():
            self.config.set(
                "log",
                "stream_stdout",
                "True",
                source="log",
                merge_settings=ConfigMergeSettings.on_change_only,
            )

        self.task_instance_cache = TaskInstanceCache()
        self.user_code_detector = UserCodeDetector.build_code_detector()
        self._autoload_modules = autoload_modules

        # will set up in __enter__
        # we can't initialize settings without having self defined as context
        # we assign real object only in _on_enter, however it's great for auto completion
        from dbnd._core.settings import DatabandSettings

        self.settings = None  # type: DatabandSettings

        self.initialized_context = False

    def _on_enter(self):
        pm.hook.dbnd_on_pre_init_context(ctx=self)
        run_user_func(config.get("core", "user_pre_init"))
        # if we are deserialized - we don't need to run this code again.
        if not self.initialized_context:
            # noinspection PyTypeChecker
            if self._module:
                load_python_module(self._module, "--module")

            module_from_config = config.get("databand", "module")
            if self._autoload_modules and module_from_config:
                load_python_module(
                    module_from_config, "config file (see [databand].module)"
                )

            # will be called from singleton context manager
            # we want to be able to catch all "new" inline airflow operators
            self.system_settings = DatabandSystemConfig()
            if self.system_settings.conf:
                self.config.set_values(
                    self.system_settings.conf, source="[databand]conf"
                )
            if self.system_settings.conf_file:
                conf_file = read_from_config_files(self.system_settings.conf_file)
                self.config.set_values(conf_file, source="[databand]conf")

            from dbnd._core.settings import DatabandSettings

            self.settings = DatabandSettings(databand_context=self)
            self.env = self.settings.get_env_config(self.system_settings.env)
            self.config.set_values(
                config_values={"task": {"task_env": self.system_settings.env}},
                source="context",
            )

            pm.hook.dbnd_on_new_context(ctx=self)

            # RUN USER SETUP FUNCTIONS
            _run_user_func(
                self.settings.core.__class__.user_driver_init,
                self.settings.core.user_driver_init,
            )

            self.task_run_env = RunInfoConfig().build_task_run_info()
            self.initialized_context = True
        else:
            # we get here if we are running at sub process that recreates the Context
            pm.hook.dbnd_on_existing_context(ctx=self)

        # we do it every time we go into databand_config
        self.configure_targets()
        self.settings.log.configure_dbnd_logging()

        _run_user_func(
            self.settings.core.__class__.user_init, self.settings.core.user_init
        )
        pm.hook.dbnd_post_enter_context(ctx=self)

    @property
    @cached()
    def tracking_store(self):
        return self.settings.core.build_tracking_store()

    @property
    @cached()
    def tracking_store_allow_errors(self):
        return self.settings.core.build_tracking_store(remove_failed_store=False)

    @property
    @cached()
    def databand_api_client(self):
        return self.settings.core.build_databand_api_client()

    def configure_targets(self):
        output_config = self.settings.output  # type: OutputConfig
        if output_config.hdf_format == "table":
            import pandas as pd
            from targets.marshalling import MARSHALERS
            from targets.marshalling.pandas import DataFrameToHdf5Table

            MARSHALERS[pd.DataFrame][FileFormat.hdf5] = DataFrameToHdf5Table()

    def _on_exit(self):
        pm.hook.dbnd_on_exit_context(ctx=self)

    def is_interactive(self):
        return self.name == "interactive"

    def __repr__(self):
        return "%s(name='%s')" % (self.__class__.__name__, self.name)

    def dbnd_run_task(
        self,
        task_or_task_name,
        run_uid=None,
        scheduled_run_info=None,
        send_heartbeat=True,
    ):
        # type: (Optional[Task,str], Optional[UUID], ScheduledRunInfo, bool) -> DatabandRun
        with new_databand_run(
            context=self,
            task_or_task_name=task_or_task_name,
            run_uid=run_uid,
            scheduled_run_info=scheduled_run_info,
            send_heartbeat=send_heartbeat,
        ) as run:  # type: DatabandRun
            run.run_driver()
            return run

    def __deepcopy__(self, memo):
        # create a copy with self.linked_to *not copied*, just referenced.
        return self

    def set_current(self, name, description=None):
        self.settings.run.name = name
        self.settings.run.description = description


def _run_user_func(param, value):
    if not value:
        return
    return run_user_func(value)


@seven.contextlib.contextmanager
def new_dbnd_context(conf=None, name=None, **kwargs):
    # type: (...) -> ContextManager[DatabandContext]

    with config(config_values=conf, source="inplace"):
        with DatabandContext.new_context(
            name=name, allow_override=True, **kwargs
        ) as dc:
            yield dc



from typing import List

from dbnd._core.configuration.config_store import _ConfigStore
from dbnd._core.configuration.config_value import ConfigValue
from dbnd._core.configuration.dbnd_config import DbndConfig
from dbnd._core.configuration.pprint_config import pformat_current_config
from targets.target_config import parse_target_config


class MultiSectionConfig(object):
    """
    Contains current config and the list of section names
    that we need to look at when we calculate the value
    """

    def __init__(self, config, sections):
        # type: (DbndConfig, List[str]) -> None
        self.config = config
        self.sections = sections

    def update_section(self, section, param_values, source):
        # we take values using names only
        defaults_store = _ConfigStore()
        for key, value in param_values:
            previous_value = self.config.get(section, key)
            if previous_value != value:
                cf = ConfigValue(value=value, source=source)
                defaults_store.set_config_value(section, key, cf)

        # we apply set on change only in the for loop, so we can optimize and not run all these code
        if defaults_store:
            self.config.set_values(config_values=defaults_store, source=source)

    def get_config_value(self, key):
        """
        If we have any override in the config -> use them!
        otherwise, return first value
        in case we have value in better section, but override in low priority section
        override wins!
        """
        best_config_value = None
        for section in self.sections:
            config_value = self.config.get_config_value(section, key)
            # first override win!
            if config_value:
                if config_value.override:
                    return config_value
                if best_config_value is None:
                    best_config_value = config_value
        return best_config_value

    def get_param_config_value(self, param_def):
        try:
            cf_value = self.get_config_value(key=param_def.name)
            if cf_value:
                return cf_value

            if param_def.config_path:
                return self.config.get_config_value(
                    section=param_def.config_path.section, key=param_def.config_path.key
                )
            return None
        except Exception as ex:
            raise param_def.parameter_exception("read configuration value", ex)

    @property
    def config_log(self):
        return "config for sections({config_sections}): {config}".format(
            config=pformat_current_config(
                self.config, sections=self.sections, as_table=True
            ),
            config_sections=self.sections,
        )

    def get_section(self, section_name):
        return self.config.config_layer.config.get(section_name)

    @property
    def layer(self):
        return self.config.config_layer

    def set_values(
        self, config_values, source=None, override=False, merge_settings=None
    ):
        self.config.set_values(config_values, source, override, merge_settings)

    def iter_sections(self):
        for section_name in self.sections:
            section = self.get_section(section_name)
            if section:
                yield section_name, section



from __future__ import with_statement

import os
import sys
import subprocess
import tempfile
import codecs # dump

from unittest import TestCase

import hjson

class TestAssets(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestAssets, self).__init__(*args, **kwargs)
        self.assetsDir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "assets")
        self.assets = self.load('testlist.txt', False).split('\n')
        self.maxDiff = None
        self.verma, self.vermi = sys.version_info[0:2]

    def load(self, name, cr):
        name = os.path.join(self.assetsDir, name)
        with open(name, 'rb') as f:
            text = f.read().decode('utf-8')
            text = text.replace('\r', '')
            if cr: text = text.replace('\n', '\r\n')
            return text

    def check(self, name, file, inputCr):
        text = self.load(file, inputCr)
        shouldFail = name[0:4] == "fail"

        try:
            data = hjson.loads(text)
            self.assertFalse(shouldFail, file)

            text1 = hjson.dumpsJSON(data)
            hjson1 = hjson.dumps(data, ensure_ascii=False);
            result = hjson.loads(self.load(name + "_result.json", inputCr))
            text2 = hjson.dumpsJSON(result)
            hjson2 = self.load(name + "_result.hjson", False)

            # dbg
            # with open(name + "_dbg1.txt", "w") as tmp: tmp.write(hjson1.encode("utf-8"))
            # with open(name + "_dbg2.txt", "w") as tmp: tmp.write(hjson2.encode("utf-8"))
            # with codecs.open(name + "_dbg3.txt", 'w', 'utf-8') as tmp: hjson.dump(data, tmp)

            if self.verma>2 or self.vermi>6:
                # final check fails on py2.6 because of string formatting issues
                self.assertEqual(text2, text1, file)
                self.assertEqual(hjson2, hjson1, file)

        except hjson.HjsonDecodeError as e:
            if not shouldFail:
                self.fail("raised error on parsing %s: %r" % (file, e))

    def test_files(self):
        for file in self.assets:
            name, sep, ext = file.partition("_test.")
            if name.startswith("stringify/quotes") or \
                name.startswith("extra/"): continue # ignore/not supported

            self.check(name, file, True)
            self.check(name, file, False)
